#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Volume V2 Volume action implementations"""

import argparse
import functools
import logging
from collections.abc import Iterable, Sequence
from typing import Any

from cliff import columns as cliff_columns
from openstack.block_storage.v2 import volume as _volume
from openstack import exceptions as sdk_exceptions
from openstack import utils as sdk_utils
from osc_lib.cli import format_columns
from osc_lib.cli import parseractions
from osc_lib import exceptions
from osc_lib import utils

from openstackclient.api import volume_v2
from openstackclient import command
from openstackclient.common import pagination
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common


LOG = logging.getLogger(__name__)


class KeyValueHintAction(argparse.Action):
    """Uses KeyValueAction or KeyValueAppendAction based on the given key"""

    APPEND_KEYS = ('same_host', 'different_host')

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._key_value_action = parseractions.KeyValueAction(*args, **kwargs)
        self._key_value_append_action = parseractions.KeyValueAppendAction(
            *args, **kwargs
        )
        super().__init__(*args, **kwargs)

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Any,
        option_string: str | None = None,
    ) -> None:
        if values.startswith(self.APPEND_KEYS):
            self._key_value_append_action(
                parser, namespace, values, option_string=option_string
            )
        else:
            self._key_value_action(
                parser, namespace, values, option_string=option_string
            )


class AttachmentsColumn(cliff_columns.FormattableColumn[list[Any]]):
    """Formattable column for attachments column.

    Unlike the parent FormattableColumn class, the initializer of the
    class takes server_cache as the second argument.
    osc_lib.utils.get_item_properties instantiate cliff FormattableColumn
    object with a single parameter "column value", so you need to pass
    a partially initialized class like
    ``functools.partial(AttachmentsColumn, server_cache)``.
    """

    def __init__(
        self, value: list[Any], server_cache: dict[str, Any] | None = None
    ) -> None:
        super().__init__(value)
        self._server_cache = server_cache or {}

    def human_readable(self) -> str:
        """Return a formatted string of a volume's attached instances

        :rtype: a string of formatted instances
        """

        msg = ''
        for attachment in self._value:
            server = attachment['server_id']
            if server in self._server_cache.keys():
                server = self._server_cache[server].name
            device = attachment['device']
            msg += f'Attached to {server} on {device} '
        return msg


def _format_volume(volume: _volume.Volume) -> dict[str, object]:
    # Some columns returned by openstacksdk should not be shown because they're
    # either irrelevant or duplicates
    ignored_columns = {
        # computed columns
        'location',
        # create-only columns
        'OS-SCH-HNT:scheduler_hints',
        'imageRef',
        # unnecessary columns
        'links',
    }
    optional_columns = {
        # only present if part of a consistency group
        'consistencygroup_id',
        # only present if there are image properties associated
        'volume_image_metadata',
    }

    info = volume.to_dict(original_names=True)
    data = {}
    for key, value in info.items():
        if key in ignored_columns:
            continue

        if key in optional_columns:
            if info[key] is None:
                continue

        data[key] = value

    data.update(
        {
            'properties': format_columns.DictColumn(data.pop('metadata')),
            'type': data.pop('volume_type'),
        }
    )

    return data


class CreateVolume(command.ShowOne):
    _description = _("Create new volume")

    @staticmethod
    def _check_size_arg(args: argparse.Namespace) -> None:
        """Check whether --size option is required or not.

        Require size parameter only in case when snapshot or source
        volume is not specified.
        """

        if (args.snapshot or args.source) is None and args.size is None:
            msg = _(
                "--size is a required option if --snapshot or --source are "
                "not specified"
            )
            raise exceptions.CommandError(msg)

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "name",
            metavar="<name>",
            nargs="?",
            help=_("Volume name"),
        )
        parser.add_argument(
            "--size",
            metavar="<size>",
            type=int,
            help=_(
                "Volume size in GB (required unless --snapshot or "
                "--source specified)"
            ),
        )
        parser.add_argument(
            "--type",
            metavar="<volume-type>",
            help=_("Set the type of volume"),
        )
        source_group = parser.add_mutually_exclusive_group()
        source_group.add_argument(
            "--image",
            metavar="<image>",
            help=_("Use <image> as source of volume (name or ID)"),
        )
        source_group.add_argument(
            "--snapshot",
            metavar="<snapshot>",
            help=_("Use <snapshot> as source of volume (name or ID)"),
        )
        source_group.add_argument(
            "--source",
            metavar="<volume>",
            help=_("Volume to clone (name or ID)"),
        )
        source_group.add_argument(
            "--source-replicated",
            metavar="<replicated-volume>",
            help=argparse.SUPPRESS,
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            help=_("Volume description"),
        )
        parser.add_argument(
            "--availability-zone",
            metavar="<availability-zone>",
            help=_("Create volume in <availability-zone>"),
        )
        parser.add_argument(
            "--consistency-group",
            metavar="consistency-group>",
            help=_("Consistency group where the new volume belongs to"),
        )
        parser.add_argument(
            "--property",
            metavar="<key=value>",
            action=parseractions.KeyValueAction,
            dest="properties",
            help=_(
                "Set a property to this volume "
                "(repeat option to set multiple properties)"
            ),
        )
        parser.add_argument(
            "--hint",
            metavar="<key=value>",
            action=KeyValueHintAction,
            help=_(
                "Arbitrary scheduler hint key-value pairs to help creating "
                "a volume. Repeat the option to set multiple hints. "
                "'same_host' and 'different_host' get values appended when "
                "repeated, all other keys take the last given value"
            ),
        )
        bootable_group = parser.add_mutually_exclusive_group()
        bootable_group.add_argument(
            "--bootable",
            action="store_true",
            dest="bootable",
            default=None,
            help=_("Mark volume as bootable"),
        )
        bootable_group.add_argument(
            "--non-bootable",
            action="store_false",
            dest="bootable",
            default=None,
            help=_("Mark volume as non-bootable (default)"),
        )
        readonly_group = parser.add_mutually_exclusive_group()
        readonly_group.add_argument(
            "--read-only",
            action="store_true",
            dest="read_only",
            default=None,
            help=_("Set volume to read-only access mode"),
        )
        readonly_group.add_argument(
            "--read-write",
            action="store_false",
            dest="read_only",
            default=None,
            help=_("Set volume to read-write access mode (default)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        self._check_size_arg(parsed_args)
        # size is validated in the above call to
        # _check_size_arg where we check that size
        # should be passed if we are not creating a
        # volume from snapshot or source volume
        size = parsed_args.size

        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '2'
        )
        image_client = self.app.client_manager.image

        source_volume = None
        if parsed_args.source:
            source_volume_obj = volume_client.find_volume(
                parsed_args.source, ignore_missing=False
            )
            source_volume = source_volume_obj.id
            size = max(size or 0, source_volume_obj.size)

        consistency_group = None
        if parsed_args.consistency_group:
            consistency_group = volume_v2.find_consistency_group(
                volume_client, parsed_args.consistency_group
            )['id']

        image = None
        if parsed_args.image:
            image = image_client.find_image(
                parsed_args.image, ignore_missing=False
            ).id

        snapshot = None
        if parsed_args.snapshot:
            snapshot_obj = volume_client.find_snapshot(
                parsed_args.snapshot, ignore_missing=False
            )
            snapshot = snapshot_obj.id
            # Cinder requires a value for size when creating a volume
            # even if creating from a snapshot. Cinder will create the
            # volume with at least the same size as the snapshot anyway,
            # so since we have the object here, just override the size
            # value if it's either not given or is smaller than the
            # snapshot size.
            size = max(size or 0, snapshot_obj.size)

        volume = volume_client.create_volume(
            size=size,
            snapshot_id=snapshot,
            name=parsed_args.name,
            description=parsed_args.description,
            volume_type=parsed_args.type,
            availability_zone=parsed_args.availability_zone,
            metadata=parsed_args.properties,
            image_id=image,
            source_volume_id=source_volume,
            consistency_group_id=consistency_group,
            scheduler_hints=parsed_args.hint,
        )

        if parsed_args.bootable is not None:
            try:
                if utils.wait_for_status(
                    volume_client.get_volume,
                    volume.id,
                    success_status=['available'],
                    error_status=['error'],
                    sleep_time=1,
                ):
                    volume_client.set_volume_bootable_status(
                        volume, parsed_args.bootable
                    )
                else:
                    msg = _(
                        "Volume status is not available for setting boot state"
                    )
                    raise exceptions.CommandError(msg)
            except Exception as e:
                LOG.error(_("Failed to set volume bootable property: %s"), e)

        if parsed_args.read_only is not None:
            try:
                if utils.wait_for_status(
                    volume_client.get_volume,
                    volume.id,
                    success_status=['available'],
                    error_status=['error'],
                    sleep_time=1,
                ):
                    volume_client.set_volume_readonly(
                        volume, parsed_args.read_only
                    )
                else:
                    msg = _(
                        "Volume status is not available for setting it"
                        "read only."
                    )
                    raise exceptions.CommandError(msg)
            except Exception as e:
                LOG.error(
                    _("Failed to set volume read-only access mode flag: %s"),
                    e,
                )

        data = _format_volume(volume)
        col_headers, col_data = zip(*sorted(data.items()))
        return col_headers, col_data


class DeleteVolume(command.Command):
    _description = _("Delete volume(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "volumes",
            metavar="<volume>",
            nargs="+",
            help=_("Volume(s) to delete (name or ID)"),
        )
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            "--force",
            action="store_true",
            help=_(
                "Attempt forced removal of volume(s), regardless of state "
                "(defaults to False)"
            ),
        )
        group.add_argument(
            "--cascade",
            action="store_true",
            help=_(
                "Remove any snapshots along with volume(s) (defaults to False)"
            ),
        )
        group.add_argument(
            # now called "cascade", accept old arg for compatibility
            "--purge",
            action="store_true",
            help=argparse.SUPPRESS,
            dest='cascade',
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '2'
        )
        result = 0

        for volume in parsed_args.volumes:
            try:
                volume_obj = volume_client.find_volume(
                    volume, ignore_missing=False
                )
                volume_client.delete_volume(
                    volume_obj.id,
                    force=parsed_args.force,
                    cascade=parsed_args.cascade,
                )
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete volume with "
                        "name or ID '%(volume)s': %(e)s"
                    ),
                    {'volume': volume, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.volumes)
            msg = _("%(result)s of %(total)s volumes failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListVolume(command.Lister):
    _description = _("List volumes")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_('Filter results by project (name or ID) (admin only)'),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        parser.add_argument(
            '--user',
            metavar='<user>',
            help=_('Filter results by user (name or ID) (admin only)'),
        )
        identity_common.add_user_domain_option_to_parser(parser)
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('Filter results by volume name'),
        )
        parser.add_argument(
            '--status',
            metavar='<status>',
            help=_('Filter results by status'),
        )
        parser.add_argument(
            '--all-projects',
            action='store_true',
            default=False,
            help=_('Include all projects (admin only)'),
        )
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_('List additional fields in output'),
        )
        pagination.add_marker_pagination_option_to_parser(
            parser, include_max_items=False
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '2'
        )
        # we don't narrow types here since we want to handle both v2.0 and v3
        identity_client = self.app.client_manager.sdk_connection.identity

        columns: tuple[str, ...] = ('id', 'name', 'status', 'size')
        column_headers: tuple[str, ...] = ('ID', 'Name', 'Status', 'Size')
        if parsed_args.long:
            columns += ('volume_type', 'is_bootable')
            column_headers += ('Type', 'Bootable')
        columns += ('attachments',)
        column_headers += ('Attached to',)
        if parsed_args.long:
            columns += ('metadata',)
            column_headers += ('Properties',)

        kwargs = {}

        if parsed_args.name:
            kwargs['name'] = parsed_args.name

        if parsed_args.status:
            kwargs['status'] = parsed_args.status

        if parsed_args.limit:
            kwargs['limit'] = parsed_args.limit

        if parsed_args.marker:
            kwargs['marker'] = parsed_args.marker

        if parsed_args.project:
            project_id = identity_common.find_project_id_sdk(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            )
            kwargs['project_id'] = project_id

        user_id = None
        if parsed_args.user:
            user_id = identity_common.find_user_id_sdk(
                identity_client, parsed_args.user, parsed_args.user_domain
            )
            kwargs['user_id'] = user_id

        # set value of 'all_tenants' when using project option
        all_projects = bool(parsed_args.project) or parsed_args.all_projects
        kwargs['all_projects'] = all_projects

        data = list(volume_client.volumes(**kwargs))

        do_server_list = False

        for vol in data:
            if vol.status == 'in-use':
                do_server_list = True
                break

        # Cache the server list
        server_cache = {}
        if do_server_list:
            try:
                compute_client = self.app.client_manager.compute
                for s in compute_client.servers():
                    server_cache[s.id] = s
            except sdk_exceptions.SDKException:
                # Just forget it if there's any trouble
                pass

        AttachmentsColumnWithCache = functools.partial(
            AttachmentsColumn, server_cache=server_cache
        )

        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={
                        'metadata': format_columns.DictColumn,
                        'attachments': AttachmentsColumnWithCache,
                    },
                )
                for s in data
            ),
        )


class MigrateVolume(command.Command):
    _description = _("Migrate volume to a new host")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar="<volume>",
            help=_("Volume to migrate (name or ID)"),
        )
        parser.add_argument(
            '--host',
            metavar="<host>",
            required=True,
            help=_(
                "Destination host (takes the form: host@backend-name#pool)"
            ),
        )
        parser.add_argument(
            '--force-host-copy',
            action="store_true",
            help=_(
                "Enable generic host-based force-migration, "
                "which bypasses driver optimizations"
            ),
        )
        parser.add_argument(
            '--lock-volume',
            action="store_true",
            help=_(
                "If specified, the volume state will be locked "
                "and will not allow a migration to be aborted "
                "(possibly by another operation)"
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '2'
        )

        volume = volume_client.find_volume(
            parsed_args.volume, ignore_missing=False
        )
        volume_client.migrate_volume(
            volume.id,
            host=parsed_args.host,
            force_host_copy=parsed_args.force_host_copy,
            lock_volume=parsed_args.lock_volume,
        )


class SetVolume(command.Command):
    _description = _("Set volume properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar='<volume>',
            help=_('Volume to modify (name or ID)'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('New volume name'),
        )
        parser.add_argument(
            '--size',
            metavar='<size>',
            type=int,
            help=_('Extend volume size in GB'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('New volume description'),
        )
        parser.add_argument(
            "--no-property",
            dest="no_property",
            action="store_true",
            help=_(
                "Remove all properties from <volume> "
                "(specify both --no-property and --property to "
                "remove the current properties before setting "
                "new properties.)"
            ),
        )
        parser.add_argument(
            '--property',
            metavar='<key=value>',
            action=parseractions.KeyValueAction,
            dest="properties",
            help=_(
                'Set a property on this volume '
                '(repeat option to set multiple properties)'
            ),
        )
        parser.add_argument(
            '--image-property',
            metavar='<key=value>',
            action=parseractions.KeyValueAction,
            dest="image_properties",
            help=_(
                'Set an image property on this volume '
                '(repeat option to set multiple image properties)'
            ),
        )
        parser.add_argument(
            "--state",
            metavar="<state>",
            choices=[
                'available',
                'error',
                'creating',
                'deleting',
                'in-use',
                'attaching',
                'detaching',
                'error_deleting',
                'maintenance',
            ],
            help=_(
                'New volume state ("available", "error", "creating", '
                '"deleting", "in-use", "attaching", "detaching", '
                '"error_deleting" or "maintenance") (admin only) '
                '(This option simply changes the state of the volume '
                'in the database with no regard to actual status, '
                'exercise caution when using)'
            ),
        )
        attached_group = parser.add_mutually_exclusive_group()
        attached_group.add_argument(
            "--attached",
            action="store_true",
            help=_(
                'Set volume attachment status to "attached" '
                '(admin only) '
                '(This option simply changes the state of the volume '
                'in the database with no regard to actual status, '
                'exercise caution when using)'
            ),
        )
        attached_group.add_argument(
            "--detached",
            action="store_true",
            help=_(
                'Set volume attachment status to "detached" '
                '(admin only) '
                '(This option simply changes the state of the volume '
                'in the database with no regard to actual status, '
                'exercise caution when using)'
            ),
        )
        parser.add_argument(
            '--type',
            metavar='<volume-type>',
            help=_('New volume type (name or ID)'),
        )
        parser.add_argument(
            '--retype-policy',
            metavar='<retype-policy>',
            choices=['never', 'on-demand'],
            help=argparse.SUPPRESS,
        )
        parser.add_argument(
            '--migration-policy',
            metavar='<migration-policy>',
            choices=['never', 'on-demand'],
            help=_(
                'Migration policy while re-typing volume '
                '("never" or "on-demand", default is "never" ) '
                '(available only when --type option is specified)'
            ),
        )
        bootable_group = parser.add_mutually_exclusive_group()
        bootable_group.add_argument(
            "--bootable",
            action="store_true",
            dest="bootable",
            default=None,
            help=_("Mark volume as bootable"),
        )
        bootable_group.add_argument(
            "--non-bootable",
            action="store_false",
            dest="bootable",
            default=None,
            help=_("Mark volume as non-bootable"),
        )
        readonly_group = parser.add_mutually_exclusive_group()
        readonly_group.add_argument(
            "--read-only",
            action="store_true",
            dest="read_only",
            default=None,
            help=_("Set volume to read-only access mode"),
        )
        readonly_group.add_argument(
            "--read-write",
            action="store_false",
            dest="read_only",
            default=None,
            help=_("Set volume to read-write access mode"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '2'
        )
        volume = volume_client.find_volume(
            parsed_args.volume, ignore_missing=False
        )

        result = 0
        if parsed_args.retype_policy:
            msg = _(
                "The '--retype-policy' option has been deprecated in favor "
                "of '--migration-policy' option. The '--retype-policy' option "
                "will be removed in a future release. Please use "
                "'--migration-policy' instead."
            )
            self.log.warning(msg)

        if parsed_args.size:
            try:
                if parsed_args.size <= volume.size:
                    msg = (
                        _("New size must be greater than %d GB") % volume.size
                    )
                    raise exceptions.CommandError(msg)
                if volume.status != 'available':
                    msg = (
                        _(
                            "Volume is in %s state, it must be available "
                            "before size can be extended"
                        )
                        % volume.status
                    )
                    raise exceptions.CommandError(msg)
                volume_client.extend_volume(volume, parsed_args.size)
            except Exception as e:
                LOG.error(_("Failed to set volume size: %s"), e)
                result += 1

        if parsed_args.no_property:
            try:
                volume_client.delete_volume_metadata(volume)
            except Exception as e:
                LOG.error(_("Failed to clean volume properties: %s"), e)
                result += 1

        if parsed_args.properties:
            try:
                volume_client.set_volume_metadata(
                    volume, **parsed_args.properties
                )
            except Exception as e:
                LOG.error(_("Failed to set volume properties: %s"), e)
                result += 1

        if parsed_args.image_properties:
            try:
                volume_client.set_volume_image_metadata(
                    volume, **parsed_args.image_properties
                )
            except Exception as e:
                LOG.error(_("Failed to set image properties: %s"), e)
                result += 1

        if parsed_args.state:
            try:
                volume_client.reset_volume_status(
                    volume, status=parsed_args.state
                )
            except Exception as e:
                LOG.error(_("Failed to set volume state: %s"), e)
                result += 1

        if parsed_args.attached:
            try:
                volume_client.reset_volume_status(
                    volume, attach_status="attached"
                )
            except Exception as e:
                LOG.error(_("Failed to set volume attach-status: %s"), e)
                result += 1

        if parsed_args.detached:
            try:
                volume_client.reset_volume_status(
                    volume, attach_status="detached"
                )
            except Exception as e:
                LOG.error(_("Failed to set volume attach-status: %s"), e)
                result += 1

        if parsed_args.bootable is not None:
            try:
                volume_client.set_volume_bootable_status(
                    volume, parsed_args.bootable
                )
            except Exception as e:
                LOG.error(_("Failed to set volume bootable property: %s"), e)
                result += 1

        if parsed_args.read_only is not None:
            try:
                volume_client.set_volume_readonly(
                    volume, parsed_args.read_only
                )
            except Exception as e:
                LOG.error(
                    _("Failed to set volume read-only access mode flag: %s"),
                    e,
                )
                result += 1

        policy = parsed_args.migration_policy or parsed_args.retype_policy

        if parsed_args.type:
            # get the migration policy
            migration_policy = 'never'
            if policy:
                migration_policy = policy
            try:
                # find the volume type
                volume_type = volume_client.find_type(
                    parsed_args.type, ignore_missing=False
                )
                # reset to the new volume type
                volume_client.retype_volume(
                    volume.id, volume_type.id, migration_policy
                )
            except Exception as e:
                LOG.error(_("Failed to set volume type: %s"), e)
                result += 1
        elif policy:
            # If the "--migration-policy" is specified without "--type"
            LOG.warning(
                _("'%s' option will not work without '--type' option"),
                (
                    '--migration-policy'
                    if parsed_args.migration_policy
                    else '--retype-policy'
                ),
            )

        kwargs = {}
        if parsed_args.name:
            kwargs['display_name'] = parsed_args.name
        if parsed_args.description:
            kwargs['display_description'] = parsed_args.description
        if kwargs:
            try:
                volume_client.update_volume(volume, **kwargs)
            except Exception as e:
                LOG.error(
                    _(
                        "Failed to update volume display name "
                        "or display description: %s"
                    ),
                    e,
                )
                result += 1

        if result > 0:
            raise exceptions.CommandError(
                _("One or more of the set operations failed")
            )


class ShowVolume(command.ShowOne):
    _description = _("Display volume details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar="<volume>",
            help=_("Volume to display (name or ID)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '2'
        )

        volume = volume_client.find_volume(
            parsed_args.volume, ignore_missing=False
        )

        data = _format_volume(volume)
        col_headers, col_data = zip(*sorted(data.items()))
        return col_headers, col_data


class UnsetVolume(command.Command):
    _description = _("Unset volume properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar='<volume>',
            help=_('Volume to modify (name or ID)'),
        )
        parser.add_argument(
            '--property',
            metavar='<key>',
            action='append',
            dest='properties',
            help=_(
                'Remove a property from volume '
                '(repeat option to remove multiple properties)'
            ),
        )
        parser.add_argument(
            '--image-property',
            metavar='<key>',
            action='append',
            dest='image_properties',
            help=_(
                'Remove an image property from volume '
                '(repeat option to remove multiple image properties)'
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '2'
        )
        volume = volume_client.find_volume(
            parsed_args.volume, ignore_missing=False
        )

        result = 0
        if parsed_args.properties:
            try:
                volume_client.delete_volume_metadata(
                    volume.id, keys=parsed_args.properties
                )
            except Exception as e:
                LOG.error(_("Failed to unset volume properties: %s"), e)
                result += 1

        if parsed_args.image_properties:
            try:
                volume_client.delete_volume_image_metadata(
                    volume.id, keys=parsed_args.image_properties
                )
            except Exception as e:
                LOG.error(_("Failed to unset image properties: %s"), e)
                result += 1

        if result > 0:
            raise exceptions.CommandError(
                _("One or more of the unset operations failed")
            )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.block_storage.v3 import group as _group
from openstack.block_storage.v3 import group_snapshot as _group_snapshot
from openstack.test import fakes as sdk_fakes
from osc_lib import exceptions

from openstackclient.tests.unit.volume.v3 import fakes as volume_fakes
from openstackclient.volume.v3 import volume_group_snapshot


class TestVolumeGroupSnapshotCreate(volume_fakes.TestVolume):
    fake_volume_group = sdk_fakes.generate_fake_resource(_group.Group)
    fake_volume_group_snapshot = sdk_fakes.generate_fake_resource(
        _group_snapshot.GroupSnapshot,
    )

    columns = (
        'ID',
        'Status',
        'Name',
        'Description',
        'Group',
        'Group Type',
    )
    data = (
        fake_volume_group_snapshot.id,
        fake_volume_group_snapshot.status,
        fake_volume_group_snapshot.name,
        fake_volume_group_snapshot.description,
        fake_volume_group_snapshot.group_id,
        fake_volume_group_snapshot.group_type_id,
    )

    def setUp(self):
        super().setUp()

        self.volume_client.find_group.return_value = self.fake_volume_group
        self.volume_client.create_group_snapshot.return_value = (
            self.fake_volume_group_snapshot
        )
        self.volume_client.find_group_snapshot.return_value = (
            self.fake_volume_group_snapshot
        )

        self.cmd = volume_group_snapshot.CreateVolumeGroupSnapshot(
            self.app, None
        )

    def test_volume_group_snapshot_create(self):
        self.set_volume_api_version('3.14')

        arglist = [
            self.fake_volume_group.id,
        ]
        verifylist = [
            ('volume_group', self.fake_volume_group.id),
            ('name', None),
            ('description', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.find_group.assert_called_once_with(
            self.fake_volume_group.id,
            ignore_missing=False,
            details=False,
        )
        self.volume_client.create_group_snapshot.assert_called_once_with(
            group_id=self.fake_volume_group.id,
            name=None,
            description=None,
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_volume_group_snapshot_create_with_options(self):
        self.set_volume_api_version('3.14')

        arglist = [
            self.fake_volume_group.id,
            '--name',
            'foo',
            '--description',
            'hello, world',
        ]
        verifylist = [
            ('volume_group', self.fake_volume_group.id),
            ('name', 'foo'),
            ('description', 'hello, world'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.find_group.assert_called_once_with(
            self.fake_volume_group.id,
            ignore_missing=False,
            details=False,
        )
        self.volume_client.create_group_snapshot.assert_called_once_with(
            group_id=self.fake_volume_group.id,
            name='foo',
            description='hello, world',
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_volume_group_snapshot_create_pre_v314(self):
        self.set_volume_api_version('3.13')

        arglist = [
            self.fake_volume_group.id,
        ]
        verifylist = [
            ('volume_group', self.fake_volume_group.id),
            ('name', None),
            ('description', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )
        self.assertIn(
            '--os-volume-api-version 3.14 or greater is required',
            str(exc),
        )


class TestVolumeGroupSnapshotDelete(volume_fakes.TestVolume):
    fake_volume_group_snapshot = sdk_fakes.generate_fake_resource(
        _group_snapshot.GroupSnapshot,
    )

    def setUp(self):
        super().setUp()

        self.volume_client.find_group_snapshot.return_value = (
            self.fake_volume_group_snapshot
        )
        self.volume_client.delete_group_snapshot.return_value = None

        self.cmd = volume_group_snapshot.DeleteVolumeGroupSnapshot(
            self.app, None
        )

    def test_volume_group_snapshot_delete(self):
        self.set_volume_api_version('3.14')

        arglist = [
            self.fake_volume_group_snapshot.id,
        ]
        verifylist = [
            ('snapshot', self.fake_volume_group_snapshot.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.delete_group_snapshot.assert_called_once_with(
            self.fake_volume_group_snapshot.id,
        )
        self.assertIsNone(result)

    def test_volume_group_snapshot_delete_pre_v314(self):
        self.set_volume_api_version('3.13')

        arglist = [
            self.fake_volume_group_snapshot.id,
        ]
        verifylist = [
            ('snapshot', self.fake_volume_group_snapshot.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )
        self.assertIn(
            '--os-volume-api-version 3.14 or greater is required',
            str(exc),
        )


class TestVolumeGroupSnapshotList(volume_fakes.TestVolume):
    fake_volume_group_snapshots = list(
        sdk_fakes.generate_fake_resources(
            _group_snapshot.GroupSnapshot,
            count=3,
        )
    )

    columns = (
        'ID',
        'Status',
        'Name',
    )
    data = [
        (
            fake_volume_group_snapshot.id,
            fake_volume_group_snapshot.status,
            fake_volume_group_snapshot.name,
        )
        for fake_volume_group_snapshot in fake_volume_group_snapshots
    ]

    def setUp(self):
        super().setUp()

        self.volume_client.group_snapshots.return_value = (
            self.fake_volume_group_snapshots
        )

        self.cmd = volume_group_snapshot.ListVolumeGroupSnapshot(
            self.app, None
        )

    def test_volume_group_snapshot_list(self):
        self.set_volume_api_version('3.14')

        arglist = [
            '--all-projects',
        ]
        verifylist = [
            ('all_projects', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.group_snapshots.assert_called_once_with(
            all_projects=True,
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(tuple(self.data), data)

    def test_volume_group_snapshot_list_pre_v314(self):
        self.set_volume_api_version('3.13')

        arglist = []
        verifylist = [
            ('all_projects', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )
        self.assertIn(
            '--os-volume-api-version 3.14 or greater is required',
            str(exc),
        )



# Copyright 2017 FUJITSU LIMITED
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from openstack.identity.v3 import project as _project
from openstack.network.v2 import vpn_ike_policy as _vpn_ike_policy
from openstack.test import fakes as sdk_fakes
from osc_lib import exceptions

from openstackclient.network.v2.vpnaas import ikepolicy
from openstackclient.tests.unit.network.v2 import fakes as test_fakes
from openstackclient.tests.unit.network.v2.vpnaas import fakes
from openstackclient.tests.unit import utils as tests_utils


class TestIKEPolicy(test_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()

        self.project = sdk_fakes.generate_fake_resource(_project.Project)
        self.identity_sdk_client.find_project.return_value = self.project

        self._ikepolicy = fakes.IKEPolicy().create()

        def _mock_ikepolicy(*args, **kwargs):
            self.network_client.find_vpn_ike_policy.assert_called_once_with(
                self._ikepolicy.id, ignore_missing=False
            )
            return {'id': args[0]}

        self.network_client.find_vpn_ike_policy.side_effect = mock.Mock(
            side_effect=_mock_ikepolicy
        )
        self.headers = (
            'ID',
            'Name',
            'Authentication Algorithm',
            'Encryption Algorithm',
            'IKE Version',
            'Perfect Forward Secrecy (PFS)',
            'Description',
            'Phase1 Negotiation Mode',
            'Project',
            'Lifetime',
        )
        self.ordered_headers = (
            'Authentication Algorithm',
            'Description',
            'Encryption Algorithm',
            'ID',
            'IKE Version',
            'Lifetime',
            'Name',
            'Perfect Forward Secrecy (PFS)',
            'Phase1 Negotiation Mode',
            'Project',
        )
        self.ordered_data = (
            self._ikepolicy['auth_algorithm'],
            self._ikepolicy['description'],
            self._ikepolicy['encryption_algorithm'],
            self._ikepolicy.id,
            self._ikepolicy['ike_version'],
            self._ikepolicy['lifetime'],
            self._ikepolicy['name'],
            self._ikepolicy['pfs'],
            self._ikepolicy['phase1_negotiation_mode'],
            self._ikepolicy['project_id'],
        )
        self.ordered_columns = (
            'auth_algorithm',
            'description',
            'encryption_algorithm',
            'id',
            'ike_version',
            'lifetime',
            'name',
            'pfs',
            'phase1_negotiation_mode',
            'project_id',
        )


class TestCreateIKEPolicy(TestIKEPolicy):
    def setUp(self):
        super().setUp()

        self.project = sdk_fakes.generate_fake_resource(_project.Project)
        self.identity_sdk_client.find_project.return_value = self.project
        self.network_client.create_vpn_ike_policy.return_value = (
            self._ikepolicy
        )

        self.cmd = ikepolicy.CreateIKEPolicy(self.app, None)

    def test_create_with_no_options(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_create_with_all_params(self):
        self.network_client.create_vpn_ike_policy.return_value = (
            self._ikepolicy
        )

        arglist = [
            '--description',
            'my-desc',
            '--auth-algorithm',
            'sha1',
            '--encryption-algorithm',
            'aes-128',
            '--phase1-negotiation-mode',
            'main',
            '--ike-version',
            'v1',
            '--pfs',
            'group5',
            '--project',
            self.project.id,
            self._ikepolicy.name,
        ]
        verifylist = [
            ('description', 'my-desc'),
            ('auth_algorithm', 'sha1'),
            ('encryption_algorithm', 'aes-128'),
            ('phase1_negotiation_mode', 'main'),
            ('ike_version', 'v1'),
            ('pfs', 'group5'),
            ('project', self.project.id),
            ('name', self._ikepolicy.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.ordered_data = tuple(
            self._ikepolicy[column] for column in self.ordered_columns
        )
        self.network_client.create_vpn_ike_policy.assert_called_once_with(
            description='my-desc',
            auth_algorithm='sha1',
            encryption_algorithm='aes-128',
            phase1_negotiation_mode='main',
            ike_version='v1',
            pfs='group5',
            project_id=self.project.id,
            name=self._ikepolicy.name,
        )
        self.assertEqual(self.ordered_headers, tuple(sorted(headers)))
        self.assertEqual(self.ordered_data, data)


class TestDeleteIKEPolicy(TestIKEPolicy):
    def setUp(self):
        super().setUp()

        self.cmd = ikepolicy.DeleteIKEPolicy(self.app, None)

    def test_delete_with_one_resource(self):
        def _mock_ike_p(*args, **kwargs):
            return _vpn_ike_policy.VpnIkePolicy(**{'id': args[0]})

        self.network_client.find_vpn_ike_policy.side_effect = _mock_ike_p

        arglist = [self._ikepolicy.id]
        verifylist = [('ikepolicy', [self._ikepolicy.id])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.delete_vpn_ike_policy.assert_called_once_with(
            self._ikepolicy.id
        )
        self.assertIsNone(result)

    def test_delete_with_multiple_resources(self):

        def _mock_ike_p(*args, **kwargs):
            return _vpn_ike_policy.VpnIkePolicy(**{'id': args[0]})

        self.network_client.find_vpn_ike_policy.side_effect = _mock_ike_p

        arglist = ['target1', 'target2']
        verifylist = [('ikepolicy', ['target1', 'target2'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.assertEqual(
            2, self.network_client.delete_vpn_ike_policy.call_count
        )
        for idx, reference in enumerate(['target1', 'target2']):
            actual = ''.join(
                self.network_client.delete_vpn_ike_policy.call_args_list[idx][
                    0
                ]
            )
            self.assertEqual(reference, actual)

    def test_delete_multiple_with_exception(self):
        arglist = ['target1']
        verifylist = [('ikepolicy', ['target1'])]

        self.network_client.find_vpn_ike_policy.side_effect = [
            'target1',
            exceptions.CommandError,
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )


class TestListIKEPolicy(TestIKEPolicy):
    def setUp(self):
        super().setUp()

        self.cmd = ikepolicy.ListIKEPolicy(self.app, None)

        self.short_header = (
            'ID',
            'Name',
            'Authentication Algorithm',
            'Encryption Algorithm',
            'IKE Version',
            'Perfect Forward Secrecy (PFS)',
        )

        self.short_data = (
            self._ikepolicy.id,
            self._ikepolicy['name'],
            self._ikepolicy['auth_algorithm'],
            self._ikepolicy['encryption_algorithm'],
            self._ikepolicy['ike_version'],
            self._ikepolicy['pfs'],
        )

        self.network_client.vpn_ike_policies.return_value = [self._ikepolicy]

    def test_list_with_long_option(self):
        arglist = ['--long']
        verifylist = [('long', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, _data = self.cmd.take_action(parsed_args)

        self.network_client.vpn_ike_policies.assert_called_once_with()
        self.assertEqual(list(self.headers), headers)

    def test_list_with_no_option(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.network_client.vpn_ike_policies.assert_called_once_with()
        self.assertEqual(list(self.short_header), headers)
        self.assertEqual([self.short_data], list(data))


class TestSetIKEPolicy(TestIKEPolicy):
    def setUp(self):
        super().setUp()

        self.network_client.update_vpn_ike_policy.return_value = (
            self._ikepolicy
        )

        self.cmd = ikepolicy.SetIKEPolicy(self.app, None)

    def test_set_auth_algorithm_with_sha256(self):
        arglist = [
            self._ikepolicy.id,
            '--auth-algorithm',
            'sha256',
        ]
        verifylist = [
            ('ikepolicy', self._ikepolicy.id),
            ('auth_algorithm', 'sha256'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_vpn_ike_policy.assert_called_once_with(
            self._ikepolicy.id, **{'auth_algorithm': 'sha256'}
        )
        self.assertIsNone(result)

    def test_set_phase1_negotiation_mode_with_aggressive(self):
        arglist = [
            self._ikepolicy.id,
            '--phase1-negotiation-mode',
            'aggressive',
        ]
        verifylist = [
            ('ikepolicy', self._ikepolicy.id),
            ('phase1_negotiation_mode', 'aggressive'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_vpn_ike_policy.assert_called_once_with(
            self._ikepolicy.id, **{'phase1_negotiation_mode': 'aggressive'}
        )
        self.assertIsNone(result)

    def test_set_name(self):
        arglist = [
            self._ikepolicy.id,
            '--name',
            'foo',
        ]
        verifylist = [
            ('ikepolicy', self._ikepolicy.id),
            ('name', 'foo'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_vpn_ike_policy.assert_called_once_with(
            self._ikepolicy.id, name='foo'
        )
        self.assertIsNone(result)

    def test_set_description(self):
        arglist = [
            self._ikepolicy.id,
            '--description',
            'foo',
        ]
        verifylist = [
            ('ikepolicy', self._ikepolicy.id),
            ('description', 'foo'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_vpn_ike_policy.assert_called_once_with(
            self._ikepolicy.id, description='foo'
        )
        self.assertIsNone(result)


class TestShowIKEPolicy(TestIKEPolicy):
    def setUp(self):
        super().setUp()

        self.network_client.find_vpn_ike_policy.side_effect = None
        self.network_client.find_vpn_ike_policy.return_value = self._ikepolicy

        self.cmd = ikepolicy.ShowIKEPolicy(self.app, None)

    def test_show_filtered_by_id_or_name(self):
        arglist = [self._ikepolicy.id]
        verifylist = [('ikepolicy', self._ikepolicy.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.network_client.find_vpn_ike_policy.assert_called_once_with(
            self._ikepolicy.id, ignore_missing=False
        )
        self.assertEqual(self.ordered_headers, headers)
        self.assertCountEqual(self.ordered_data, data)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.block_storage.v3 import group_type as _group_type
from openstack.test import fakes as sdk_fakes
from osc_lib.cli import format_columns
from osc_lib import exceptions

from openstackclient.tests.unit.volume.v3 import fakes as volume_fakes
from openstackclient.volume.v3 import volume_group_type


class TestVolumeGroupType(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()
        self.set_volume_api_version('3.11')


class TestVolumeGroupTypeCreate(TestVolumeGroupType):
    fake_volume_group_type = sdk_fakes.generate_fake_resource(
        _group_type.GroupType
    )

    columns = (
        'ID',
        'Name',
        'Description',
        'Is Public',
        'Properties',
    )
    data = (
        fake_volume_group_type.id,
        fake_volume_group_type.name,
        fake_volume_group_type.description,
        fake_volume_group_type.is_public,
        format_columns.DictColumn(fake_volume_group_type.group_specs),
    )

    def setUp(self):
        super().setUp()

        self.volume_client.create_group_type.return_value = (
            self.fake_volume_group_type
        )

        self.cmd = volume_group_type.CreateVolumeGroupType(self.app, None)

    def test_volume_group_type_create(self):
        arglist = [
            self.fake_volume_group_type.name,
        ]
        verifylist = [
            ('name', self.fake_volume_group_type.name),
            ('description', None),
            ('is_public', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.create_group_type.assert_called_once_with(
            name=self.fake_volume_group_type.name,
            description=None,
            is_public=True,
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_volume_group_type_create_with_options(self):
        arglist = [
            self.fake_volume_group_type.name,
            '--description',
            'foo',
            '--private',
        ]
        verifylist = [
            ('name', self.fake_volume_group_type.name),
            ('description', 'foo'),
            ('is_public', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.create_group_type.assert_called_once_with(
            name=self.fake_volume_group_type.name,
            description='foo',
            is_public=False,
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_volume_group_type_create_pre_v311(self):
        self.set_volume_api_version('3.10')

        arglist = [
            self.fake_volume_group_type.name,
        ]
        verifylist = [
            ('name', self.fake_volume_group_type.name),
            ('description', None),
            ('is_public', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.11 or greater is required', str(exc)
        )


class TestVolumeGroupTypeDelete(TestVolumeGroupType):
    fake_volume_group_type = sdk_fakes.generate_fake_resource(
        _group_type.GroupType
    )

    def setUp(self):
        super().setUp()

        self.volume_client.find_group_type.return_value = (
            self.fake_volume_group_type
        )
        self.volume_client.delete_group_type.return_value = None

        self.cmd = volume_group_type.DeleteVolumeGroupType(self.app, None)

    def test_volume_group_type_delete(self):
        arglist = [
            self.fake_volume_group_type.id,
        ]
        verifylist = [
            ('group_type', self.fake_volume_group_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.find_group_type.assert_called_once_with(
            self.fake_volume_group_type.id,
            ignore_missing=False,
        )
        self.volume_client.delete_group_type.assert_called_once_with(
            self.fake_volume_group_type,
        )
        self.assertIsNone(result)

    def test_volume_group_type_delete_pre_v311(self):
        self.set_volume_api_version('3.10')

        arglist = [
            self.fake_volume_group_type.id,
        ]
        verifylist = [
            ('group_type', self.fake_volume_group_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.11 or greater is required', str(exc)
        )


class TestVolumeGroupTypeSet(TestVolumeGroupType):
    fake_volume_group_type = sdk_fakes.generate_fake_resource(
        _group_type.GroupType, group_specs={'foo': 'bar'}
    )

    columns = (
        'ID',
        'Name',
        'Description',
        'Is Public',
        'Properties',
    )
    data = (
        fake_volume_group_type.id,
        fake_volume_group_type.name,
        fake_volume_group_type.description,
        fake_volume_group_type.is_public,
        format_columns.DictColumn(fake_volume_group_type.group_specs),
    )

    def setUp(self):
        super().setUp()

        self.volume_client.find_group_type.return_value = (
            self.fake_volume_group_type
        )
        self.volume_client.update_group_type.return_value = (
            self.fake_volume_group_type
        )

        self.cmd = volume_group_type.SetVolumeGroupType(self.app, None)

    def test_volume_group_type_set(self):
        arglist = [
            self.fake_volume_group_type.id,
            '--name',
            'foo',
            '--description',
            'hello, world',
            '--public',
            '--property',
            'fizz=buzz',
        ]
        verifylist = [
            ('group_type', self.fake_volume_group_type.id),
            ('name', 'foo'),
            ('description', 'hello, world'),
            ('is_public', True),
            ('no_property', False),
            ('properties', {'fizz': 'buzz'}),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.update_group_type.assert_called_once_with(
            self.fake_volume_group_type.id,
            name='foo',
            description='hello, world',
            is_public=True,
        )
        self.volume_client.create_group_type_group_specs.assert_called_once_with(
            self.fake_volume_group_type.id,
            {'fizz': 'buzz'},
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_volume_group_type_with_no_property_option(self):
        arglist = [
            self.fake_volume_group_type.id,
            '--no-property',
            '--property',
            'fizz=buzz',
        ]
        verifylist = [
            ('group_type', self.fake_volume_group_type.id),
            ('name', None),
            ('description', None),
            ('is_public', None),
            ('no_property', True),
            ('properties', {'fizz': 'buzz'}),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.delete_group_type_group_specs_property.assert_called_once_with(
            self.fake_volume_group_type.id,
            'foo',
        )
        self.volume_client.create_group_type_group_specs.assert_called_once_with(
            self.fake_volume_group_type.id,
            {'fizz': 'buzz'},
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_volume_group_type_set_pre_v311(self):
        self.set_volume_api_version('3.10')

        arglist = [
            self.fake_volume_group_type.id,
            '--name',
            'foo',
            '--description',
            'hello, world',
        ]
        verifylist = [
            ('group_type', self.fake_volume_group_type.id),
            ('name', 'foo'),
            ('description', 'hello, world'),
            ('is_public', None),
            ('no_property', False),
            ('properties', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.11 or greater is required', str(exc)
        )


class TestVolumeGroupTypeUnset(TestVolumeGroupType):
    fake_volume_group_type = sdk_fakes.generate_fake_resource(
        _group_type.GroupType
    )

    columns = (
        'ID',
        'Name',
        'Description',
        'Is Public',
        'Properties',
    )
    data = (
        fake_volume_group_type.id,
        fake_volume_group_type.name,
        fake_volume_group_type.description,
        fake_volume_group_type.is_public,
        format_columns.DictColumn(fake_volume_group_type.group_specs),
    )

    def setUp(self):
        super().setUp()

        self.volume_client.find_group_type.return_value = (
            self.fake_volume_group_type
        )

        self.cmd = volume_group_type.UnsetVolumeGroupType(self.app, None)

    def test_volume_group_type_unset(self):
        arglist = [
            self.fake_volume_group_type.id,
            '--property',
            'fizz',
        ]
        verifylist = [
            ('group_type', self.fake_volume_group_type.id),
            ('properties', ['fizz']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.delete_group_type_group_specs_property.assert_called_once_with(
            self.fake_volume_group_type.id,
            'fizz',
        )
        self.volume_client.find_group_type.assert_called_once_with(
            self.fake_volume_group_type.id,
            ignore_missing=False,
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_volume_group_type_unset_pre_v311(self):
        self.set_volume_api_version('3.10')

        arglist = [
            self.fake_volume_group_type.id,
            '--property',
            'fizz',
        ]
        verifylist = [
            ('group_type', self.fake_volume_group_type.id),
            ('properties', ['fizz']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.11 or greater is required', str(exc)
        )


class TestVolumeGroupTypeList(TestVolumeGroupType):
    fake_volume_group_types = list(
        sdk_fakes.generate_fake_resources(_group_type.GroupType, count=2)
    )

    columns = (
        'ID',
        'Name',
        'Is Public',
        'Group Specs',
    )
    data = [
        (
            fake_volume_group_type.id,
            fake_volume_group_type.name,
            fake_volume_group_type.is_public,
            format_columns.DictColumn(fake_volume_group_type.group_specs),
        )
        for fake_volume_group_type in fake_volume_group_types
    ]

    def setUp(self):
        super().setUp()

        self.volume_client.group_types.return_value = (
            self.fake_volume_group_types
        )
        self.volume_client.get_group_type.return_value = (
            self.fake_volume_group_types[0]
        )

        self.cmd = volume_group_type.ListVolumeGroupType(self.app, None)

    def test_volume_group_type_list(self):
        arglist = []
        verifylist = [
            ('show_default', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.group_types.assert_called_once_with()
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(tuple(self.data), data)

    def test_volume_group_type_list_with_default_option(self):
        arglist = [
            '--default',
        ]
        verifylist = [
            ('show_default', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.get_group_type.assert_called_once_with('default')
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(
            tuple([self.data[0]]),
            data,
        )

    def test_volume_group_type_list_pre_v311(self):
        self.set_volume_api_version('3.10')

        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.11 or greater is required', str(exc)
        )


class TestVolumeGroupTypeShow(TestVolumeGroupType):
    fake_volume_group_type = sdk_fakes.generate_fake_resource(
        _group_type.GroupType
    )

    columns = (
        'ID',
        'Name',
        'Description',
        'Is Public',
        'Properties',
    )
    data = (
        fake_volume_group_type.id,
        fake_volume_group_type.name,
        fake_volume_group_type.description,
        fake_volume_group_type.is_public,
        format_columns.DictColumn(fake_volume_group_type.group_specs),
    )

    def setUp(self):
        super().setUp()

        self.volume_client.find_group_type.return_value = (
            self.fake_volume_group_type
        )

        self.cmd = volume_group_type.ShowVolumeGroupType(self.app, None)

    def test_volume_group_type_show(self):
        arglist = [
            self.fake_volume_group_type.id,
        ]
        verifylist = [
            ('group_type', self.fake_volume_group_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.find_group_type.assert_called_once_with(
            self.fake_volume_group_type.id,
            ignore_missing=False,
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_volume_group_type_show_pre_v311(self):
        self.set_volume_api_version('3.10')

        arglist = [
            self.fake_volume_group_type.id,
        ]
        verifylist = [
            ('group_type', self.fake_volume_group_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.11 or greater is required', str(exc)
        )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from manilaclient import api_versions
from osc_lib import exceptions
from osc_lib import utils as oscutils

from openstackclient.share.v2 import messages
from openstackclient.tests.unit.share.v2 import fakes as share_fakes
from openstackclient.tests.unit import utils as test_utils

COLUMNS = [
    'ID',
    'Resource Type',
    'Resource ID',
    'Action ID',
    'User Message',
    'Detail ID',
    'Created At',
]


class TestMessage(share_fakes.TestShare):
    def setUp(self):
        super().setUp()

        self.messages_mock = self.share_client.messages
        self.messages_mock.reset_mock()

        self.set_share_api_version(api_versions.MAX_VERSION)


class TestMessageDelete(TestMessage):
    def setUp(self):
        super().setUp()

        self.message = share_fakes.FakeMessage.create_one_message()

        self.messages_mock.get.return_value = self.message

        self.cmd = messages.DeleteMessage(self.app, None)

    def test_message_delete_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_message_delete(self):
        arglist = [self.message.id]
        verifylist = [('message', [self.message.id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.messages_mock.delete.assert_called_with(self.message)
        self.assertIsNone(result)

    def test_message_delete_multiple(self):
        messages = share_fakes.FakeMessage.create_messages(count=2)
        arglist = [messages[0].id, messages[1].id]
        verifylist = [('message', [messages[0].id, messages[1].id])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertEqual(self.messages_mock.delete.call_count, len(messages))
        self.assertIsNone(result)

    def test_message_delete_exception(self):
        arglist = [self.message.id]
        verifylist = [('message', [self.message.id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.messages_mock.delete.side_effect = exceptions.CommandError()
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestMessageShow(TestMessage):
    def setUp(self):
        super().setUp()

        self.message = share_fakes.FakeMessage.create_one_message()
        self.messages_mock.get.return_value = self.message

        self.cmd = messages.ShowMessage(self.app, None)

        self.data = self.message._info.values()
        self.columns = self.message._info.keys()

    def test_message_show_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_message_show(self):
        arglist = [self.message.id]
        verifylist = [('message', self.message.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.messages_mock.get.assert_called_with(self.message.id)
        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)


class TestMessageList(TestMessage):
    def setUp(self):
        super().setUp()

        self.messages = share_fakes.FakeMessage.create_messages(count=2)

        self.messages_mock.list.return_value = self.messages

        self.values = (
            oscutils.get_dict_properties(m._info, COLUMNS)
            for m in self.messages
        )

        self.cmd = messages.ListMessage(self.app, None)

    def test_list_messages(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.messages_mock.list.assert_called_with(
            search_opts={
                'limit': None,
                'request_id': None,
                'resource_type': None,
                'resource_id': None,
                'action_id': None,
                'detail_id': None,
                'message_level': None,
                'created_since': None,
                'created_before': None,
            }
        )

        self.assertEqual(COLUMNS, columns)
        self.assertEqual(list(self.values), list(data))

    def test_list_messages_api_version_exception(self):
        self.set_share_api_version('2.50')

        arglist = [
            '--before',
            '2021-02-06T09:49:58-05:00',
            '--since',
            '2021-02-05T09:49:58-05:00',
        ]
        verifylist = [
            ('before', '2021-02-06T09:49:58-05:00'),
            ('since', '2021-02-05T09:49:58-05:00'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import copy
from unittest import mock

from osc_lib import exceptions

from openstackclient.identity.v2_0 import role_assignment
from openstackclient.tests.unit import fakes
from openstackclient.tests.unit.identity.v2_0 import fakes as identity_fakes


class TestRoleAssignment(identity_fakes.TestIdentity):
    def setUp(self):
        super().setUp()


class TestRoleAssignmentList(TestRoleAssignment):
    columns = (
        'Role',
        'User',
        'Project',
    )

    def setUp(self):
        super(TestRoleAssignment, self).setUp()

        # Get a shortcut to the UserManager Mock
        self.users_mock = self.identity_client.users
        self.users_mock.reset_mock()

        # Get a shortcut to the ProjectManager Mock
        self.projects_mock = self.identity_client.projects
        self.projects_mock.reset_mock()

        # Get a shortcut to the RoleManager Mock
        self.roles_mock = self.identity_client.roles
        self.roles_mock.reset_mock()

        self.projects_mock.get.return_value = fakes.FakeResource(
            None,
            copy.deepcopy(identity_fakes.PROJECT),
            loaded=True,
        )

        self.users_mock.get.return_value = fakes.FakeResource(
            None,
            copy.deepcopy(identity_fakes.USER),
            loaded=True,
        )

        self.roles_mock.roles_for_user.return_value = [
            fakes.FakeResource(
                None,
                copy.deepcopy(identity_fakes.ROLE),
                loaded=True,
            ),
        ]

        # Get the command object to test
        self.cmd = role_assignment.ListRoleAssignment(self.app, None)

    def test_role_assignment_list_no_filters(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # This argument combination should raise a CommandError
        self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )

    def test_role_assignment_list_only_project_filter(self):
        arglist = [
            '--project',
            identity_fakes.project_name,
        ]
        verifylist = [
            ('project', identity_fakes.project_name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # This argument combination should raise a CommandError
        self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )

    def test_role_assignment_list_only_user_filter(self):
        arglist = [
            '--user',
            identity_fakes.user_name,
        ]
        verifylist = [
            ('user', identity_fakes.user_name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # This argument combination should raise a CommandError
        self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )

    def test_role_assignment_list_project_and_user(self):
        self.roles_mock.roles_for_user.return_value = [
            fakes.FakeResource(
                None,
                copy.deepcopy(identity_fakes.ROLE),
                loaded=True,
            ),
            fakes.FakeResource(
                None,
                copy.deepcopy(identity_fakes.ROLE_2),
                loaded=True,
            ),
        ]

        arglist = [
            '--project',
            identity_fakes.project_name,
            '--user',
            identity_fakes.user_name,
        ]
        verifylist = [
            ('user', identity_fakes.user_name),
            ('project', identity_fakes.project_name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.roles_mock.roles_for_user.assert_called_with(
            identity_fakes.user_id,
            identity_fakes.project_id,
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                identity_fakes.role_id,
                identity_fakes.user_id,
                identity_fakes.project_id,
            ),
            (
                identity_fakes.ROLE_2['id'],
                identity_fakes.user_id,
                identity_fakes.project_id,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_role_assignment_list_def_creds(self):
        self.app.client_manager.auth_ref = mock.Mock()
        auth_ref = self.app.client_manager.auth_ref
        auth_ref.project_id.return_value = identity_fakes.project_id
        auth_ref.user_id.return_value = identity_fakes.user_id

        self.roles_mock.roles_for_user.return_value = [
            fakes.FakeResource(
                None,
                copy.deepcopy(identity_fakes.ROLE),
                loaded=True,
            ),
            fakes.FakeResource(
                None,
                copy.deepcopy(identity_fakes.ROLE_2),
                loaded=True,
            ),
        ]

        arglist = [
            '--auth-user',
            '--auth-project',
        ]
        verifylist = [
            ('authuser', True),
            ('authproject', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.roles_mock.roles_for_user.assert_called_with(
            identity_fakes.user_id,
            identity_fakes.project_id,
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                identity_fakes.role_id,
                identity_fakes.user_id,
                identity_fakes.project_id,
            ),
            (
                identity_fakes.ROLE_2['id'],
                identity_fakes.user_id,
                identity_fakes.project_id,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_role_assignment_list_by_name_project_and_user(self):
        self.roles_mock.roles_for_user.return_value = [
            fakes.FakeResource(
                None,
                copy.deepcopy(identity_fakes.ROLE),
                loaded=True,
            ),
            fakes.FakeResource(
                None,
                copy.deepcopy(identity_fakes.ROLE_2),
                loaded=True,
            ),
        ]

        arglist = [
            '--project',
            identity_fakes.project_name,
            '--user',
            identity_fakes.user_name,
            '--names',
        ]
        verifylist = [
            ('user', identity_fakes.user_name),
            ('project', identity_fakes.project_name),
            ('names', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.roles_mock.roles_for_user.assert_called_with(
            identity_fakes.user_id,
            identity_fakes.project_id,
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                identity_fakes.role_name,
                identity_fakes.user_name,
                identity_fakes.project_name,
            ),
            (
                identity_fakes.ROLE_2['name'],
                identity_fakes.user_name,
                identity_fakes.project_name,
            ),
        )
        self.assertEqual(datalist, tuple(data))



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest.mock import call

from openstack.block_storage.v3 import transfer as _transfer
from openstack.block_storage.v3 import volume as _volume
from openstack import exceptions as sdk_exceptions
from openstack.test import fakes as sdk_fakes
from osc_lib import exceptions

from openstackclient.tests.unit import utils as test_utils
from openstackclient.tests.unit.volume.v3 import fakes as volume_fakes
from openstackclient.volume.v3 import volume_transfer_request


class TestTransferAccept(volume_fakes.TestVolume):
    columns = (
        'id',
        'name',
        'volume_id',
    )

    def setUp(self):
        super().setUp()

        self.volume_transfer = sdk_fakes.generate_fake_resource(
            _transfer.Transfer
        )
        self.data = (
            self.volume_transfer.id,
            self.volume_transfer.name,
            self.volume_transfer.volume_id,
        )

        self.volume_client.find_transfer.return_value = self.volume_transfer
        self.volume_client.accept_transfer.return_value = self.volume_transfer

        self.cmd = volume_transfer_request.AcceptTransferRequest(
            self.app, None
        )

    def test_transfer_accept(self):
        arglist = [
            '--auth-key',
            'key_value',
            self.volume_transfer.id,
        ]
        verifylist = [
            ('transfer_request', self.volume_transfer.id),
            ('auth_key', 'key_value'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.find_transfer.assert_called_once_with(
            self.volume_transfer.id, ignore_missing=False
        )
        self.volume_client.accept_transfer.assert_called_once_with(
            self.volume_transfer.id,
            'key_value',
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_transfer_accept_non_admin(self):
        """Non-admin users get ResourceNotFound on find_transfer; we fall back."""
        self.volume_client.find_transfer.side_effect = (
            sdk_exceptions.ResourceNotFound
        )
        arglist = [
            '--auth-key',
            'key_value',
            self.volume_transfer.id,
        ]
        verifylist = [
            ('transfer_request', self.volume_transfer.id),
            ('auth_key', 'key_value'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.accept_transfer.assert_called_once_with(
            self.volume_transfer.id,
            'key_value',
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_transfer_accept_no_option(self):
        arglist = [self.volume_transfer.id]
        verifylist = [('transfer_request', self.volume_transfer.id)]

        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )


class TestTransferCreate(volume_fakes.TestVolume):
    columns = (
        'auth_key',
        'created_at',
        'id',
        'name',
        'volume_id',
    )

    def setUp(self):
        super().setUp()

        self.volume = sdk_fakes.generate_fake_resource(_volume.Volume)
        self.volume_transfer = sdk_fakes.generate_fake_resource(
            _transfer.Transfer, volume_id=self.volume.id
        )
        self.data = (
            self.volume_transfer.auth_key,
            self.volume_transfer.created_at,
            self.volume_transfer.id,
            self.volume_transfer.name,
            self.volume_transfer.volume_id,
        )

        self.volume_client.find_volume.return_value = self.volume
        self.volume_client.create_transfer.return_value = self.volume_transfer

        self.cmd = volume_transfer_request.CreateTransferRequest(
            self.app, None
        )

    def test_transfer_create_without_name(self):
        arglist = [self.volume.id]
        verifylist = [('volume', self.volume.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.create_transfer.assert_called_once_with(
            volume_id=self.volume.id,
            name=None,
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_transfer_create_with_name(self):
        arglist = [
            '--name',
            self.volume_transfer.name,
            self.volume.id,
        ]
        verifylist = [
            ('name', self.volume_transfer.name),
            ('volume', self.volume.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.create_transfer.assert_called_once_with(
            volume_id=self.volume.id,
            name=self.volume_transfer.name,
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_transfer_create_with_no_snapshots(self):
        self.set_volume_api_version('3.55')

        arglist = [
            '--no-snapshots',
            self.volume.id,
        ]
        verifylist = [
            ('name', None),
            ('snapshots', False),
            ('volume', self.volume.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.create_transfer.assert_called_once_with(
            volume_id=self.volume.id,
            name=None,
            no_snapshots=True,
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_transfer_create_pre_v355(self):
        self.set_volume_api_version('3.54')

        arglist = [
            '--no-snapshots',
            self.volume.id,
        ]
        verifylist = [
            ('name', None),
            ('snapshots', False),
            ('volume', self.volume.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.55 or greater is required', str(exc)
        )


class TestTransferDelete(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.volume_transfers = [
            sdk_fakes.generate_fake_resource(_transfer.Transfer),
            sdk_fakes.generate_fake_resource(_transfer.Transfer),
        ]
        self.volume_client.delete_transfer.return_value = None

        self.cmd = volume_transfer_request.DeleteTransferRequest(
            self.app, None
        )

    def test_transfer_delete(self):
        arglist = [self.volume_transfers[0].id]
        verifylist = [("transfer_request", [self.volume_transfers[0].id])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.delete_transfer.assert_called_once_with(
            self.volume_transfers[0].id, ignore_missing=False
        )
        self.assertIsNone(result)

    def test_delete_multiple_transfers(self):
        arglist = [v.id for v in self.volume_transfers]
        verifylist = [('transfer_request', arglist)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        calls = [
            call(v.id, ignore_missing=False) for v in self.volume_transfers
        ]
        self.volume_client.delete_transfer.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_delete_multiple_transfers_with_exception(self):
        arglist = [
            self.volume_transfers[0].id,
            'unexist_transfer',
        ]
        verifylist = [('transfer_request', arglist)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.volume_client.delete_transfer.side_effect = [
            None,
            exceptions.CommandError,
        ]

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual(
                '1 of 2 volume transfer requests failed to delete',
                str(e),
            )


class TestTransferList(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.volume_transfers = [
            sdk_fakes.generate_fake_resource(_transfer.Transfer),
        ]
        self.volume_client.transfers.return_value = self.volume_transfers

        self.cmd = volume_transfer_request.ListTransferRequest(self.app, None)

    def test_transfer_list_without_argument(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.transfers.assert_called_once_with(
            details=True, all_projects=False
        )
        self.assertEqual(('ID', 'Name', 'Volume'), columns)
        self.assertEqual(
            (
                (
                    self.volume_transfers[0].id,
                    self.volume_transfers[0].name,
                    self.volume_transfers[0].volume_id,
                ),
            ),
            tuple(data),
        )

    def test_transfer_list_with_argument(self):
        arglist = ["--all-projects"]
        verifylist = [("all_projects", True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, _data = self.cmd.take_action(parsed_args)

        self.volume_client.transfers.assert_called_once_with(
            details=True, all_projects=True
        )
        self.assertEqual(('ID', 'Name', 'Volume'), columns)


class TestTransferShow(volume_fakes.TestVolume):
    columns = (
        'created_at',
        'id',
        'name',
        'volume_id',
    )

    def setUp(self):
        super().setUp()

        self.volume_transfer = sdk_fakes.generate_fake_resource(
            _transfer.Transfer
        )
        self.data = (
            self.volume_transfer.created_at,
            self.volume_transfer.id,
            self.volume_transfer.name,
            self.volume_transfer.volume_id,
        )

        self.volume_client.find_transfer.return_value = self.volume_transfer

        self.cmd = volume_transfer_request.ShowTransferRequest(self.app, None)

    def test_transfer_show(self):
        arglist = [self.volume_transfer.id]
        verifylist = [('transfer_request', self.volume_transfer.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.find_transfer.assert_called_once_with(
            self.volume_transfer.id, ignore_missing=False
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)



#   Copyright 2015 iWeb Technologies Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest.mock import call

from openstack.block_storage.v2 import qos_spec as _qos_spec
from openstack.block_storage.v2 import type as _type
from openstack.test import fakes as sdk_fakes
from osc_lib.cli import format_columns
from osc_lib import exceptions

from openstackclient.tests.unit.volume.v2 import fakes as volume_fakes
from openstackclient.volume.v2 import qos_specs


class TestQosAssociate(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.volume_type = sdk_fakes.generate_fake_resource(_type.Type)
        self.qos_spec = sdk_fakes.generate_fake_resource(
            _qos_spec.QoSSpec,
            consumer='front-end',
            specs={'foo': 'bar', 'iops': '9001'},
        )

        self.volume_client.find_qos_spec.return_value = self.qos_spec
        self.volume_client.find_type.return_value = self.volume_type
        self.cmd = qos_specs.AssociateQos(self.app, None)

    def test_qos_associate(self):
        arglist = [self.qos_spec.id, self.volume_type.id]
        verifylist = [
            ('qos_spec', self.qos_spec.id),
            ('volume_type', self.volume_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.find_qos_spec.assert_called_once_with(
            self.qos_spec.id, ignore_missing=False
        )
        self.volume_client.find_type.assert_called_once_with(
            self.volume_type.id, ignore_missing=False
        )
        self.volume_client.associate_qos_spec.assert_called_once_with(
            self.qos_spec.id, self.volume_type.id
        )
        self.assertIsNone(result)


class TestQosCreate(volume_fakes.TestVolume):
    columns = ('consumer', 'id', 'name', 'properties')

    def setUp(self):
        super().setUp()

        self.new_qos_spec = sdk_fakes.generate_fake_resource(
            _qos_spec.QoSSpec,
            consumer='front-end',
            specs={'foo': 'bar', 'iops': '9001'},
        )
        self.volume_client.create_qos_spec.return_value = self.new_qos_spec

        self.data = (
            self.new_qos_spec.consumer,
            self.new_qos_spec.id,
            self.new_qos_spec.name,
            format_columns.DictColumn(self.new_qos_spec.specs),
        )

        self.cmd = qos_specs.CreateQos(self.app, None)

    def test_qos_create_without_properties(self):
        arglist = [
            self.new_qos_spec.name,
        ]
        verifylist = [
            ('name', self.new_qos_spec.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.create_qos_spec.assert_called_once_with(
            name=self.new_qos_spec.name, consumer='both'
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_qos_create_with_consumer(self):
        arglist = [
            '--consumer',
            self.new_qos_spec.consumer,
            self.new_qos_spec.name,
        ]
        verifylist = [
            ('consumer', self.new_qos_spec.consumer),
            ('name', self.new_qos_spec.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.create_qos_spec.assert_called_once_with(
            name=self.new_qos_spec.name,
            consumer=self.new_qos_spec.consumer,
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_qos_create_with_properties(self):
        arglist = [
            '--consumer',
            self.new_qos_spec.consumer,
            '--property',
            'foo=bar',
            '--property',
            'iops=9001',
            self.new_qos_spec.name,
        ]
        verifylist = [
            ('consumer', self.new_qos_spec.consumer),
            ('properties', self.new_qos_spec.specs),
            ('name', self.new_qos_spec.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_client.create_qos_spec.assert_called_once_with(
            name=self.new_qos_spec.name,
            consumer=self.new_qos_spec.consumer,
            foo='bar',
            iops='9001',
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)


class TestQosDelete(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.qos_specs = [
            sdk_fakes.generate_fake_resource(_qos_spec.QoSSpec, specs={}),
            sdk_fakes.generate_fake_resource(_qos_spec.QoSSpec, specs={}),
        ]
        qos_by_id = {q.id: q for q in self.qos_specs}
        self.volume_client.find_qos_spec.side_effect = lambda x, **kwargs: (
            qos_by_id[x]
        )
        self.cmd = qos_specs.DeleteQos(self.app, None)

    def test_qos_delete(self):
        arglist = [self.qos_specs[0].id]
        verifylist = [('qos_specs', [self.qos_specs[0].id])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.delete_qos_spec.assert_called_once_with(
            self.qos_specs[0].id, ignore_missing=False, force=False
        )
        self.assertIsNone(result)

    def test_qos_delete_with_force(self):
        arglist = ['--force', self.qos_specs[0].id]
        verifylist = [('force', True), ('qos_specs', [self.qos_specs[0].id])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.delete_qos_spec.assert_called_once_with(
            self.qos_specs[0].id, ignore_missing=False, force=True
        )
        self.assertIsNone(result)

    def test_delete_multiple_qoses(self):
        arglist = []
        for q in self.qos_specs:
            arglist.append(q.id)
        verifylist = [
            ('qos_specs', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        calls = []
        for q in self.qos_specs:
            calls.append(call(q.id, ignore_missing=False, force=False))
        self.volume_client.delete_qos_spec.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_delete_multiple_qoses_with_exception(self):
        arglist = [
            self.qos_specs[0].id,
            'unexist_qos',
        ]
        verifylist = [
            ('qos_specs', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual(
                '1 of 2 QoS specifications failed to delete.', str(e)
            )

        self.volume_client.delete_qos_spec.assert_called_once_with(
            self.qos_specs[0].id, ignore_missing=False, force=False
        )


class TestQosDisassociate(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.volume_type = sdk_fakes.generate_fake_resource(_type.Type)
        self.qos_spec = sdk_fakes.generate_fake_resource(
            _qos_spec.QoSSpec, consumer='front-end', specs={}
        )

        self.volume_client.find_qos_spec.return_value = self.qos_spec
        self.volume_client.find_type.return_value = self.volume_type
        self.cmd = qos_specs.DisassociateQos(self.app, None)

    def test_qos_disassociate_with_volume_type(self):
        arglist = [
            '--volume-type',
            self.volume_type.id,
            self.qos_spec.id,
        ]
        verifylist = [
            ('volume_type', self.volume_type.id),
            ('qos_spec', self.qos_spec.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.disassociate_qos_spec.assert_called_once_with(
            self.qos_spec.id, self.volume_type.id
        )
        self.assertIsNone(result)

    def test_qos_disassociate_with_all_volume_types(self):
        arglist = [
            '--all',
            self.qos_spec.id,
        ]
        verifylist = [('qos_spec', self.qos_spec.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.disassociate_all_qos_spec.assert_called_once_with(
            self.qos_spec.id
        )
        self.assertIsNone(result)


class TestQosList(volume_fakes.TestVolume):
    columns = (
        'ID',
        'Name',
        'Consumer',
        'Associations',
        'Properties',
    )

    def setUp(self):
        super().setUp()

        self.qos_specs = [
            sdk_fakes.generate_fake_resource(
                _qos_spec.QoSSpec,
                consumer='front-end',
                specs={'foo': 'bar', 'iops': '9001'},
            ),
            sdk_fakes.generate_fake_resource(
                _qos_spec.QoSSpec,
                consumer='front-end',
                specs={'foo': 'bar', 'iops': '9001'},
            ),
        ]
        self.qos_association = sdk_fakes.generate_fake_resource(
            _qos_spec.QoSSpecAssociation
        )
        self.volume_client.qos_specs.return_value = self.qos_specs
        self.volume_client.qos_spec_associations.return_value = [
            self.qos_association
        ]

        self.data = []
        for q in self.qos_specs:
            self.data.append(
                (
                    q.id,
                    q.name,
                    q.consumer,
                    format_columns.ListColumn([self.qos_association.name]),
                    format_columns.DictColumn(q.specs),
                )
            )

        self.cmd = qos_specs.ListQos(self.app, None)

    def test_qos_list(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_client.qos_specs.assert_called_once_with()

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))


class TestQosSet(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.qos_spec = sdk_fakes.generate_fake_resource(
            _qos_spec.QoSSpec,
            consumer='front-end',
            specs={'foo': 'bar', 'iops': '9001'},
        )

        self.volume_client.find_qos_spec.return_value = self.qos_spec
        self.cmd = qos_specs.SetQos(self.app, None)

    def test_qos_set_with_properties_with_id(self):
        arglist = [
            '--no-property',
            '--property',
            'a=b',
            '--property',
            'c=d',
            self.qos_spec.id,
        ]
        new_property = {"a": "b", "c": "d"}
        verifylist = [
            ('no_property', True),
            ('properties', new_property),
            ('qos_spec', self.qos_spec.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.delete_qos_spec_metadata.assert_called_once_with(
            self.qos_spec.id,
            list(self.qos_spec.specs.keys()),
        )
        self.volume_client.update_qos_spec.assert_called_once_with(
            self.qos_spec.id, a='b', c='d'
        )
        self.assertIsNone(result)


class TestQosShow(volume_fakes.TestVolume):
    columns = ('associations', 'consumer', 'id', 'name', 'properties')

    def setUp(self):
        super().setUp()

        self.qos_spec = sdk_fakes.generate_fake_resource(
            _qos_spec.QoSSpec,
            consumer='front-end',
            specs={'foo': 'bar', 'iops': '9001'},
        )
        self.qos_association = sdk_fakes.generate_fake_resource(
            _qos_spec.QoSSpecAssociation
        )
        self.volume_client.find_qos_spec.return_value = self.qos_spec
        self.volume_client.qos_spec_associations.return_value = [
            self.qos_association
        ]

        self.data = (
            format_columns.ListColumn([self.qos_association.name]),
            self.qos_spec.consumer,
            self.qos_spec.id,
            self.qos_spec.name,
            format_columns.DictColumn(self.qos_spec.specs),
        )

        self.cmd = qos_specs.ShowQos(self.app, None)

    def test_qos_show(self):
        arglist = [self.qos_spec.id]
        verifylist = [('qos_spec', self.qos_spec.id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_client.find_qos_spec.assert_called_once_with(
            self.qos_spec.id, ignore_missing=False
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, tuple(data))


class TestQosUnset(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.qos_spec = sdk_fakes.generate_fake_resource(
            _qos_spec.QoSSpec, consumer='front-end', specs={}
        )
        self.volume_client.find_qos_spec.return_value = self.qos_spec
        self.cmd = qos_specs.UnsetQos(self.app, None)

    def test_qos_unset_with_properties(self):
        arglist = [
            '--property',
            'iops',
            '--property',
            'foo',
            self.qos_spec.id,
        ]
        verifylist = [
            ('properties', ['iops', 'foo']),
            ('qos_spec', self.qos_spec.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_client.delete_qos_spec_metadata.assert_called_once_with(
            self.qos_spec.id, ['iops', 'foo']
        )
        self.assertIsNone(result)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#
import ddt
from manilaclient import api_versions
from openstackclient.share.v2 import security_services
from osc_lib import exceptions
from osc_lib import utils as oscutils

from openstackclient.tests.unit.share.v2 import fakes as share_fakes
from openstackclient.tests.unit import utils as test_utils


class TestShareSecurityService(share_fakes.TestShare):
    def setUp(self):
        super().setUp()

        self.security_services_mock = self.share_client.security_services
        self.security_services_mock.reset_mock()

        self.share_networks_mock = self.share_client.share_networks
        self.share_networks_mock.reset_mock()

        self.set_share_api_version(api_versions.MAX_VERSION)


@ddt.ddt
class TestShareSecurityServiceCreate(TestShareSecurityService):
    def setUp(self):
        super().setUp()

        self.security_service = (
            share_fakes.FakeShareSecurityService.create_fake_security_service()
        )
        self.security_services_mock.create.return_value = self.security_service
        self.cmd = security_services.CreateShareSecurityService(self.app, None)

        self.data = self.security_service._info.values()
        self.columns = self.security_service._info.keys()

    def test_share_security_service_create_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_security_service_create(self):
        arglist = [
            self.security_service.type,
            '--dns-ip',
            self.security_service.dns_ip,
            '--ou',
            self.security_service.ou,
            '--server',
            self.security_service.server,
            '--domain',
            self.security_service.domain,
            '--user',
            self.security_service.user,
            '--password',
            self.security_service.password,
            '--name',
            self.security_service.name,
            '--description',
            self.security_service.description,
            '--default-ad-site',
            self.security_service.default_ad_site,
        ]
        verifylist = [
            ('type', self.security_service.type),
            ('dns_ip', self.security_service.dns_ip),
            ('ou', self.security_service.ou),
            ('server', self.security_service.server),
            ('domain', self.security_service.domain),
            ('user', self.security_service.user),
            ('password', self.security_service.password),
            ('name', self.security_service.name),
            ('description', self.security_service.description),
            ('default_ad_site', self.security_service.default_ad_site),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.security_services_mock.create.assert_called_with(
            self.security_service.type,
            dns_ip=self.security_service.dns_ip,
            server=self.security_service.server,
            domain=self.security_service.domain,
            user=self.security_service.user,
            password=self.security_service.password,
            name=self.security_service.name,
            description=self.security_service.description,
            ou=self.security_service.ou,
            default_ad_site=self.security_service.default_ad_site,
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    @ddt.data('2.43', '2.75')
    def test_share_security_service_create_api_version_exception(
        self, version
    ):
        self.set_share_api_version(version)

        arglist = [
            self.security_service.type,
        ]
        verifylist = [
            ('type', self.security_service.type),
        ]

        if api_versions.APIVersion(version) <= api_versions.APIVersion("2.43"):
            arglist.extend(['--ou', self.security_service.ou])
            verifylist.append(('ou', self.security_service.ou))

        if api_versions.APIVersion(version) <= api_versions.APIVersion("2.75"):
            arglist.extend(
                ['--default-ad-site', self.security_service.default_ad_site]
            )
            verifylist.append(
                ('default_ad_site', self.security_service.default_ad_site)
            )

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestShareSecurityServiceDelete(TestShareSecurityService):
    def setUp(self):
        super().setUp()

        self.security_service = (
            share_fakes.FakeShareSecurityService.create_fake_security_service()
        )
        self.security_services_mock.get.return_value = self.security_service

        self.security_services = share_fakes.FakeShareSecurityService.create_fake_security_services()

        self.cmd = security_services.DeleteShareSecurityService(self.app, None)

    def test_share_security_service_delete_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_security_service_delete(self):
        arglist = [
            self.security_services[0].id,
            self.security_services[1].id,
        ]
        verifylist = [
            (
                'security_service',
                [self.security_services[0].id, self.security_services[1].id],
            ),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertEqual(
            self.security_services_mock.delete.call_count,
            len(self.security_services),
        )
        self.assertIsNone(result)

    def test_share_security_service_delete_exception(self):
        arglist = [
            self.security_services[0].id,
        ]
        verifylist = [
            ('security_service', [self.security_services[0].id]),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.security_services_mock.delete.side_effect = (
            exceptions.CommandError()
        )
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestShareSecurityServiceShow(TestShareSecurityService):
    def setUp(self):
        super().setUp()

        self.security_service = (
            share_fakes.FakeShareSecurityService.create_fake_security_service()
        )
        self.security_services_mock.get.return_value = self.security_service

        self.cmd = security_services.ShowShareSecurityService(self.app, None)

        self.data = self.security_service._info.values()
        self.columns = self.security_service._info.keys()

    def test_share_security_service_show_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_security_service_show(self):
        arglist = [self.security_service.id]
        verifylist = [('security_service', self.security_service.id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.security_services_mock.get.assert_called_with(
            self.security_service.id
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)


@ddt.ddt
class TestShareSecurityServiceSet(TestShareSecurityService):
    def setUp(self):
        super().setUp()

        self.security_service = (
            share_fakes.FakeShareSecurityService.create_fake_security_service(
                methods={'update': None}
            )
        )
        self.security_services_mock.get.return_value = self.security_service
        self.cmd = security_services.SetShareSecurityService(self.app, None)

    def test_share_security_service_set_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_security_service_set(self):
        arglist = [
            self.security_service.id,
            '--dns-ip',
            self.security_service.dns_ip,
            '--ou',
            self.security_service.ou,
            '--server',
            self.security_service.server,
            '--domain',
            self.security_service.domain,
            '--user',
            self.security_service.user,
            '--password',
            self.security_service.password,
            '--name',
            self.security_service.name,
            '--description',
            self.security_service.description,
            '--default-ad-site',
            self.security_service.default_ad_site,
        ]
        verifylist = [
            ('security_service', self.security_service.id),
            ('dns_ip', self.security_service.dns_ip),
            ('ou', self.security_service.ou),
            ('server', self.security_service.server),
            ('domain', self.security_service.domain),
            ('user', self.security_service.user),
            ('password', self.security_service.password),
            ('name', self.security_service.name),
            ('description', self.security_service.description),
            ('default_ad_site', self.security_service.default_ad_site),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.security_service.update.assert_called_with(
            dns_ip=self.security_service.dns_ip,
            server=self.security_service.server,
            domain=self.security_service.domain,
            user=self.security_service.user,
            password=self.security_service.password,
            name=self.security_service.name,
            description=self.security_service.description,
            ou=self.security_service.ou,
            default_ad_site=self.security_service.default_ad_site,
        )
        self.assertIsNone(result)

    def test_share_security_service_set_exception(self):
        arglist = [
            self.security_service.id,
            '--name',
            self.security_service.name,
        ]
        verifylist = [
            ('security_service', self.security_service.id),
            ('name', self.security_service.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.security_service.update.side_effect = exceptions.CommandError()
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    @ddt.data('2.43', '2.75')
    def test_share_security_service_set_api_version_exception(self, version):
        self.set_share_api_version(version)

        arglist = [
            self.security_service.id,
        ]
        verifylist = [
            ('security_service', self.security_service.id),
        ]

        if api_versions.APIVersion(version) <= api_versions.APIVersion("2.43"):
            arglist.extend(['--ou', self.security_service.ou])
            verifylist.append(('ou', self.security_service.ou))

        if api_versions.APIVersion(version) <= api_versions.APIVersion("2.75"):
            arglist.extend(
                ['--default-ad-site', self.security_service.default_ad_site]
            )
            verifylist.append(
                ('default_ad_site', self.security_service.default_ad_site)
            )

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


@ddt.ddt
class TestShareSecurityServiceUnset(TestShareSecurityService):
    def setUp(self):
        super().setUp()

        self.security_service = (
            share_fakes.FakeShareSecurityService.create_fake_security_service(
                methods={'update': None}
            )
        )
        self.security_services_mock.get.return_value = self.security_service
        self.cmd = security_services.UnsetShareSecurityService(self.app, None)

    def test_share_security_service_unset_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_security_service_unset(self):
        arglist = [
            self.security_service.id,
            '--dns-ip',
            '--ou',
            '--server',
            '--domain',
            '--user',
            '--password',
            '--name',
            '--description',
            '--default-ad-site',
        ]
        verifylist = [
            ('security_service', self.security_service.id),
            ('dns_ip', True),
            ('ou', True),
            ('server', True),
            ('domain', True),
            ('user', True),
            ('password', True),
            ('name', True),
            ('description', True),
            ('default_ad_site', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.security_service.update.assert_called_with(
            dns_ip='',
            server='',
            domain='',
            user='',
            password='',
            name='',
            description='',
            ou='',
            default_ad_site='',
        )
        self.assertIsNone(result)

    def test_share_security_service_unset_exception(self):
        arglist = [
            self.security_service.id,
            '--name',
        ]
        verifylist = [
            ('security_service', self.security_service.id),
            ('name', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.security_service.update.side_effect = exceptions.CommandError()
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    @ddt.data('2.43', '2.75')
    def test_share_security_service_unset_api_version_exception(self, version):
        self.set_share_api_version(version)

        arglist = [
            self.security_service.id,
        ]
        verifylist = [
            ('security_service', self.security_service.id),
        ]

        if api_versions.APIVersion(version) <= api_versions.APIVersion("2.43"):
            arglist.extend(['--ou'])
            verifylist.append(('ou', True))

        if api_versions.APIVersion(version) <= api_versions.APIVersion("2.75"):
            (arglist.extend(['--default-ad-site']),)
            verifylist.append(('default_ad_site', True))

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestShareSecurityServiceList(TestShareSecurityService):
    columns = [
        'ID',
        'Name',
        'Status',
        'Type',
    ]

    def setUp(self):
        super().setUp()

        self.share_network = (
            share_fakes.FakeShareNetwork.create_one_share_network()
        )
        self.share_networks_mock.get.return_value = self.share_network
        self.services_list = share_fakes.FakeShareSecurityService.create_fake_security_services()
        self.security_services_mock.list.return_value = self.services_list
        self.values = (
            oscutils.get_dict_properties(i._info, self.columns)
            for i in self.services_list
        )

        self.cmd = security_services.ListShareSecurityService(self.app, None)

    def test_share_security_service_list_no_args(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.security_services_mock.list.assert_called_with(
            search_opts={
                'all_tenants': False,
                'status': None,
                'name': None,
                'type': None,
                'user': None,
                'dns_ip': None,
                'server': None,
                'domain': None,
                'offset': None,
                'limit': None,
            },
            detailed=False,
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(list(self.values), list(data))

    def test_share_security_service_list(self):
        arglist = [
            '--share-network',
            self.share_network.id,
            '--status',
            self.services_list[0].status,
            '--name',
            self.services_list[0].name,
            '--type',
            self.services_list[0].type,
            '--user',
            self.services_list[0].user,
            '--dns-ip',
            self.services_list[0].dns_ip,
            '--ou',
            self.services_list[0].ou,
            '--server',
            self.services_list[0].server,
            '--domain',
            self.services_list[0].domain,
            '--default-ad-site',
            self.services_list[0].default_ad_site,
            '--limit',
            '1',
        ]
        verifylist = [
            ('share_network', self.share_network.id),
            ('status', self.services_list[0].status),
            ('name', self.services_list[0].name),
            ('type', self.services_list[0].type),
            ('user', self.services_list[0].user),
            ('dns_ip', self.services_list[0].dns_ip),
            ('ou', self.services_list[0].ou),
            ('server', self.services_list[0].server),
            ('domain', self.services_list[0].domain),
            ('default_ad_site', self.services_list[0].default_ad_site),
            ('limit', 1),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.security_services_mock.list.assert_called_with(
            search_opts={
                'all_tenants': False,
                'status': self.services_list[0].status,
                'name': self.services_list[0].name,
                'type': self.services_list[0].type,
                'user': self.services_list[0].user,
                'dns_ip': self.services_list[0].dns_ip,
                'server': self.services_list[0].server,
                'domain': self.services_list[0].domain,
                'default_ad_site': self.services_list[0].default_ad_site,
                'offset': None,
                'limit': 1,
                'ou': self.services_list[0].ou,
                'share_network_id': self.share_network.id,
            },
            detailed=False,
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(list(self.values), list(data))

    def test_share_security_service_list_ou_api_version_exception(self):
        self.set_share_api_version('2.43')

        arglist = [
            '--ou',
            self.services_list[0].ou,
        ]
        verifylist = [
            ('ou', self.services_list[0].ou),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_share_security_service_list_ad_site_api_version_exception(self):
        self.set_share_api_version('2.75')

        arglist = [
            '--default-ad-site',
            self.services_list[0].default_ad_site,
        ]
        verifylist = [
            ('default_ad_site', self.services_list[0].default_ad_site),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_share_security_service_list_detail_all_projects(self):
        arglist = ['--all-projects', '--detail']
        verifylist = [
            ('all_projects', True),
            ('detail', True),
        ]
        columns_detail = self.columns.copy()
        columns_detail.append('Project ID')
        columns_detail.append('Share Networks')

        values_detail = (
            oscutils.get_dict_properties(i._info, columns_detail)
            for i in self.services_list
        )

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.security_services_mock.list.assert_called_with(
            search_opts={
                'all_tenants': True,
                'status': None,
                'name': None,
                'type': None,
                'user': None,
                'dns_ip': None,
                'server': None,
                'domain': None,
                'offset': None,
                'limit': None,
            },
            detailed=True,
        )
        self.assertEqual(columns_detail, columns)
        self.assertEqual(list(values_detail), list(data))



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Object v1 action implementations"""

import argparse
import logging
from collections.abc import Iterable, Sequence
from typing import Any

from osc_lib.cli import format_columns
from osc_lib.cli import parseractions
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.common import pagination
from openstackclient.i18n import _


LOG = logging.getLogger(__name__)


class CreateObject(command.Lister):
    _description = _("Upload object to container")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'container',
            metavar='<container>',
            help=_('Container for new object'),
        )
        parser.add_argument(
            'objects',
            metavar='<filename>',
            nargs="+",
            help=_('Local filename(s) to upload'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_(
                'Upload a file and rename it. '
                'Can only be used when uploading a single object'
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        if parsed_args.name:
            if len(parsed_args.objects) > 1:
                msg = _(
                    'Attempting to upload multiple objects and '
                    'using --name is not permitted'
                )
                raise exceptions.CommandError(msg)
        results = []
        for obj in parsed_args.objects:
            if len(obj) > 1024:
                LOG.warning(
                    _(
                        'Object name is %s characters long, default limit'
                        ' is 1024'
                    ),
                    len(obj),
                )
            data = self.app.client_manager.object_store.object_create(
                container=parsed_args.container,
                object=obj,
                name=parsed_args.name,
            )
            results.append(data)

        columns = ("object", "container", "etag")
        return (
            columns,
            (
                utils.get_dict_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in results
            ),
        )


class DeleteObject(command.Command):
    _description = _("Delete object from container")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'container',
            metavar='<container>',
            help=_('Delete object(s) from <container>'),
        )
        parser.add_argument(
            'objects',
            metavar='<object>',
            nargs="+",
            help=_('Object(s) to delete'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        for obj in parsed_args.objects:
            self.app.client_manager.object_store.object_delete(
                container=parsed_args.container,
                object=obj,
            )


class ListObject(command.Lister):
    _description = _("List objects")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "container",
            metavar="<container>",
            help=_("Container to list"),
        )
        parser.add_argument(
            "--prefix",
            metavar="<prefix>",
            help=_("Filter list using <prefix>"),
        )
        parser.add_argument(
            "--delimiter",
            metavar="<delimiter>",
            help=_("Roll up items with <delimiter>"),
        )
        pagination.add_marker_pagination_option_to_parser(
            parser, include_max_items=False
        )
        parser.add_argument(
            "--end-marker",
            metavar="<end-marker>",
            help=_("End anchor for paging"),
        )
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_('List additional fields in output'),
        )
        parser.add_argument(
            '--all',
            action='store_true',
            default=False,
            help=_('List all objects in container (default is 10000)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        columns: tuple[str, ...] = ('Name',)
        if parsed_args.long:
            columns += ('Bytes', 'Hash', 'Content Type', 'Last Modified')

        kwargs = {}
        if parsed_args.prefix:
            kwargs['prefix'] = parsed_args.prefix
        if parsed_args.delimiter:
            kwargs['delimiter'] = parsed_args.delimiter
        if parsed_args.marker:
            kwargs['marker'] = parsed_args.marker
        if parsed_args.end_marker:
            kwargs['end_marker'] = parsed_args.end_marker
        if parsed_args.limit:
            kwargs['limit'] = parsed_args.limit
        if parsed_args.all:
            kwargs['full_listing'] = True

        data = self.app.client_manager.object_store.object_list(
            container=parsed_args.container, **kwargs
        )

        return (
            columns,
            (
                utils.get_dict_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in data
            ),
        )


class SaveObject(command.Command):
    _description = _("Save object locally")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--file",
            metavar="<filename>",
            help=_(
                "Destination filename (defaults to object name); using '-'"
                " as the filename will print the file to stdout"
            ),
        )
        parser.add_argument(
            'container',
            metavar='<container>',
            help=_('Download <object> from <container>'),
        )
        parser.add_argument(
            "object",
            metavar="<object>",
            help=_("Object to save"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.app.client_manager.object_store.object_save(
            container=parsed_args.container,
            object=parsed_args.object,
            file=parsed_args.file,
        )


class SetObject(command.Command):
    _description = _("Set object properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'container',
            metavar='<container>',
            help=_('Modify <object> from <container>'),
        )
        parser.add_argument(
            'object',
            metavar='<object>',
            help=_('Object to modify'),
        )
        parser.add_argument(
            "--property",
            metavar="<key=value>",
            required=True,
            action=parseractions.KeyValueAction,
            help=_(
                "Set a property on this object "
                "(repeat option to set multiple properties)"
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.app.client_manager.object_store.object_set(
            parsed_args.container,
            parsed_args.object,
            properties=parsed_args.property,
        )


class ShowObject(command.ShowOne):
    _description = _("Display object details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'container',
            metavar='<container>',
            help=_('Display <object> from <container>'),
        )
        parser.add_argument(
            'object',
            metavar='<object>',
            help=_('Object to display'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        data = self.app.client_manager.object_store.object_show(
            container=parsed_args.container,
            object=parsed_args.object,
        )
        if 'properties' in data:
            data['properties'] = format_columns.DictColumn(data['properties'])

        col_headers, col_data = zip(*sorted(data.items()))
        return col_headers, col_data


class UnsetObject(command.Command):
    _description = _("Unset object properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'container',
            metavar='<container>',
            help=_('Modify <object> from <container>'),
        )
        parser.add_argument(
            'object',
            metavar='<object>',
            help=_('Object to modify'),
        )
        parser.add_argument(
            '--property',
            metavar='<key>',
            required=True,
            action='append',
            default=[],
            help=_(
                'Property to remove from object '
                '(repeat option to remove multiple properties)'
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.app.client_manager.object_store.object_unset(
            parsed_args.container,
            parsed_args.object,
            properties=parsed_args.property,
        )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any
import uuid

from manilaclient.common.apiclient import utils as apiutils
from manilaclient.common import constants
from openstackclient.identity import common as identity_common
from osc_lib import exceptions
from osc_lib import utils as osc_utils

from openstackclient import command
from openstackclient.i18n import _

LOG = logging.getLogger(__name__)

LOCK_DETAIL_ATTRIBUTES = [
    'ID',
    'Resource Id',
    'Resource Type',
    'Resource Action',
    'Lock Context',
    'User Id',
    'Project Id',
    'Created At',
    'Updated At',
    'Lock Reason',
]

LOCK_SUMMARY_ATTRIBUTES = [
    'ID',
    'Resource Id',
    'Resource Type',
    'Resource Action',
]

RESOURCE_TYPE_MANAGERS = {
    'share': 'shares',
    'access_rule': 'share_access_rules',
}


# TODO(stephenfin): Move this to osc_lib since it's useful elsewhere (e.g.
# glance)
def is_uuid_like(value: str) -> bool:
    """Returns validation of a value as a UUID.

    :param val: Value to verify
    :type val: string
    :returns: bool

    .. versionchanged:: 1.1.1
       Support non-lowercase UUIDs.
    """
    try:
        formatted_value = (
            value.replace('urn:', '')
            .replace('uuid:', '')
            .strip('{}')
            .replace('-', '')
            .lower()
        )
        return str(uuid.UUID(value)).replace('-', '') == formatted_value
    except (TypeError, ValueError, AttributeError):
        return False


class CreateResourceLock(command.ShowOne):
    """Create a new resource lock."""

    _description = _("Lock a resource action from occurring on a resource")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'resource',
            metavar='<resource_name_or_id>',
            help=_('Name or ID of resource to lock.'),
        )
        parser.add_argument(
            'resource_type',
            metavar='<resource_type>',
            help=_('Type of the resource (e.g.: share, access).'),
        )
        parser.add_argument(
            '--resource-action',
            '--resource_action',
            metavar='<resource_action>',
            default='delete',
            help=_('Action to lock on the resource (default="delete")'),
        )
        parser.add_argument(
            '--lock-reason',
            '--lock_reason',
            '--reason',
            metavar='<lock_reason>',
            help=_('Reason for the resource lock.'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], tuple[Any, ...]]:
        share_client = self.app.client_manager.share
        resource_type = parsed_args.resource_type
        if resource_type not in RESOURCE_TYPE_MANAGERS:
            raise exceptions.CommandError(_("Unsupported resource type"))
        res_manager = RESOURCE_TYPE_MANAGERS[resource_type]

        resource = osc_utils.find_resource(
            getattr(share_client, res_manager), parsed_args.resource
        )
        resource_lock = share_client.resource_locks.create(
            resource.id,
            resource_type,
            parsed_args.resource_action,
            parsed_args.lock_reason,
        )

        resource_lock._info.pop('links', None)

        return self.dict2columns(resource_lock._info)


class DeleteResourceLock(command.Command):
    """Remove one or more resource locks."""

    _description = _("Remove one or more resource locks")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'lock',
            metavar='<lock>',
            nargs='+',
            help=_('ID(s) of the lock(s).'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        share_client = self.app.client_manager.share
        failure_count = 0

        for lock in parsed_args.lock:
            try:
                lock = apiutils.find_resource(
                    share_client.resource_locks, lock
                )
                lock.delete()
            except Exception as e:
                failure_count += 1
                LOG.error(
                    _("Failed to delete %(lock)s: %(e)s"),
                    {'lock': lock, 'e': e},
                )

        if failure_count > 0:
            raise exceptions.CommandError(
                _("Unable to delete some or all of the specified locks.")
            )


class ListResourceLock(command.Lister):
    """Lists all resource locks."""

    _description = _("Lists all resource locks")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--all-projects',
            action='store_true',
            help=_("Filter resource locks for all projects. (Admin only)."),
        )
        parser.add_argument(
            '--project',
            default=None,
            help=_(
                "Filter resource locks for specific project by name or ID, "
                "combine with --all-projects (Admin only)."
            ),
        )
        parser.add_argument(
            '--user',
            default=None,
            help=_(
                "Filter resource locks for specific user by name or ID, "
                "combine with --all-projects to search across projects "
                "(Admin only)."
            ),
        )
        parser.add_argument(
            '--id',
            metavar='<id>',
            default=None,
            help=_('Filter resource locks by ID. Default=None.'),
        )
        parser.add_argument(
            '--resource',
            '--resource-id',
            '--resource_id',
            default=None,
            metavar='<resource-id>',
            dest='resource',
            help=_(
                "Filter resource locks for a resource by ID, specify "
                "--resource-type to look up by name."
            ),
        )
        parser.add_argument(
            '--resource-type',
            '--resource_type',
            default=None,
            metavar='<resource_type>',
            help=_("Filter resource locks by type of resource."),
        )
        parser.add_argument(
            '--resource-action',
            '--resource_action',
            default=None,
            metavar='<resource_action>',
            help=_("Filter resource locks by resource action."),
        )

        parser.add_argument(
            '--lock-context',
            '--lock_context',
            '--context',
            default=None,
            choices=['user', 'admin', 'service'],
            metavar='<lock_context>',
            help=_("Filter resource locks by context."),
        )
        parser.add_argument(
            '--since',
            default=None,
            metavar='<created_since>',
            help=_(
                "Filter resource locks created since given date. "
                "The date format must be conforming to ISO8601. "
            ),
        )
        parser.add_argument(
            '--before',
            default=None,
            metavar='<created_before>',
            help=_(
                "Filter resource locks created before given date. "
                "The date format must be conforming to ISO8601. "
            ),
        )
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            type=int,
            default=None,
            help=_("Number of resource locks to list. (Default=None)"),
        )
        parser.add_argument(
            '--offset',
            metavar="<offset>",
            default=None,
            help=_(
                'Starting position of resource lock records '
                'in a paginated list.'
            ),
        )
        parser.add_argument(
            '--sort-key',
            '--sort_key',
            metavar='<sort_key>',
            type=str,
            default=None,
            choices=constants.RESOURCE_LOCK_SORT_KEY_VALUES,
            help=(
                f'Key to be sorted, available keys are '
                f'{constants.RESOURCE_LOCK_SORT_KEY_VALUES}. Default=None.'
            ),
        )
        parser.add_argument(
            '--sort-dir',
            '--sort_dir',
            metavar='<sort_dir>',
            type=str,
            default=None,
            choices=constants.SORT_DIR_VALUES,
            help=(
                f'Sort direction, available values are '
                f'{constants.SORT_DIR_VALUES}. OPTIONAL: Default=None.'
            ),
        )
        parser.add_argument(
            '--detailed',
            dest='detailed',
            metavar='<0|1>',
            nargs='?',
            type=int,
            const=1,
            default=0,
            help=_("Show detailed information about filtered resource locks."),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        share_client = self.app.client_manager.share
        identity_client = self.app.client_manager.identity

        columns = (
            LOCK_SUMMARY_ATTRIBUTES
            if not parsed_args.detailed
            else LOCK_DETAIL_ATTRIBUTES
        )

        project_id = None
        user_id = None

        if parsed_args.project:
            project_id = identity_common.find_project(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            ).id
        if parsed_args.user:
            user_id = identity_common.find_user(
                identity_client, parsed_args.user, parsed_args.user_domain
            ).id
        # set all_projects when using project option
        all_projects = bool(parsed_args.project) or parsed_args.all_projects

        resource_id = parsed_args.resource
        resource_type = parsed_args.resource_type
        if resource_type is not None:
            if resource_type not in RESOURCE_TYPE_MANAGERS:
                raise exceptions.CommandError(_("Unsupported resource type"))
            if resource_id is not None:
                res_manager = RESOURCE_TYPE_MANAGERS[resource_type]
                resource_id = osc_utils.find_resource(
                    getattr(share_client, res_manager), parsed_args.resource
                ).id
        elif resource_id and not is_uuid_like(resource_id):
            raise exceptions.CommandError(
                _("Provide resource ID or specify --resource-type.")
            )

        search_opts = {
            'all_projects': all_projects,
            'project_id': project_id,
            'user_id': user_id,
            'id': parsed_args.id,
            'resource_id': resource_id,
            'resource_type': parsed_args.resource_type,
            'resource_action': parsed_args.resource_action,
            'lock_context': parsed_args.lock_context,
            'created_before': parsed_args.before,
            'created_since': parsed_args.since,
            'limit': parsed_args.limit,
            'offset': parsed_args.offset,
        }

        resource_locks = share_client.resource_locks.list(
            search_opts=search_opts,
            sort_key=parsed_args.sort_key,
            sort_dir=parsed_args.sort_dir,
        )

        return (
            columns,
            (
                osc_utils.get_item_properties(m, columns)
                for m in resource_locks
            ),
        )


class ShowResourceLock(command.ShowOne):
    """Show details about a resource lock."""

    _description = _("Show details about a resource lock")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'lock', metavar='<lock>', help=_('ID of resource lock to show.')
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], tuple[Any, ...]]:
        share_client = self.app.client_manager.share

        resource_lock = apiutils.find_resource(
            share_client.resource_locks, parsed_args.lock
        )

        return (
            LOCK_DETAIL_ATTRIBUTES,
            osc_utils.get_dict_properties(
                resource_lock._info, LOCK_DETAIL_ATTRIBUTES
            ),
        )


class SetResourceLock(command.Command):
    """Set resource lock properties."""

    _description = _("Update resource lock properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'lock', metavar='<lock>', help=_('ID of lock to update.')
        )
        parser.add_argument(
            '--resource-action',
            '--resource_action',
            metavar='<resource_action>',
            help=_('Resource action to set in the resource lock'),
        )
        parser.add_argument(
            '--lock-reason',
            '--lock_reason',
            '--reason',
            dest='lock_reason',
            help=_("Reason for the resource lock"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        share_client = self.app.client_manager.share

        update_kwargs = {}
        if parsed_args.resource_action is not None:
            update_kwargs['resource_action'] = parsed_args.resource_action
        if parsed_args.lock_reason is not None:
            update_kwargs['lock_reason'] = parsed_args.lock_reason
        if update_kwargs:
            share_client.resource_locks.update(
                parsed_args.lock, **update_kwargs
            )


class UnsetResourceLock(command.Command):
    """Unsets a property on a resource lock."""

    _description = _("Remove resource lock properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'lock', metavar='<lock>', help=_('ID of resource lock to update.')
        )
        parser.add_argument(
            '--lock-reason',
            '--lock_reason',
            '--reason',
            dest='lock_reason',
            action='store_true',
            default=False,
            help=_("Unset the lock reason. (Default=False)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        share_client = self.app.client_manager.share

        if parsed_args.lock_reason:
            share_client.resource_locks.update(
                parsed_args.lock, lock_reason=None
            )
